from pathlib import Path
from typing import Optional, Dict, List, Union, Any

from phidata.app.db import DbApp, DbAppArgs
from phidata.infra.k8s.enums.service_type import ServiceType
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class PostgresVolumeType(ExtendedEnum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"
    AwsEbs = "AwsEbs"
    # For backwards compatibility
    AWS_EBS = "AWS_EBS"


class PostgresDbArgs(DbAppArgs):
    name: str = "postgres"
    version: str = "1"
    enabled: bool = True

    # -*- Image Configuration
    image_name: str = "postgres"
    image_tag: str = "14"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # -*- Postgres Configuration
    # Provide POSTGRES_USER as db_user or POSTGRES_USER in secrets_file
    db_user: Optional[str] = None
    # Provide POSTGRES_PASSWORD as db_password or POSTGRES_PASSWORD in secrets_file
    db_password: Optional[str] = None
    # Provide POSTGRES_DB as db_schema or POSTGRES_DB in secrets_file
    db_schema: Optional[str] = None
    db_driver: str = "postgresql"
    pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata"
    postgres_initdb_args: Optional[str] = None
    postgres_initdb_waldir: Optional[str] = None
    postgres_host_auth_method: Optional[str] = None
    postgres_password_file: Optional[str] = None
    postgres_user_file: Optional[str] = None
    postgres_db_file: Optional[str] = None
    postgres_initdb_args_file: Optional[str] = None

    # -*- Container Configuration
    container_name: Optional[str] = None

    # Container ports
    # Open a container port if open_container_port=True
    open_container_port: bool = True
    # Port number on the container
    container_port: int = 5432
    # Port name: Only used by the K8sContainer
    container_port_name: str = "pg"
    # Host port: Only used by the DockerContainer
    container_host_port: int = 5432

    # Container volumes
    create_volume: bool = True
    volume_name: Optional[str] = None
    volume_type: PostgresVolumeType = PostgresVolumeType.EmptyDir
    # Container path to mount the postgres volume
    # should be the parent directory of pgdata
    volume_container_path: str = "/var/lib/postgresql/data"
    # Host path to mount the postgres volume
    # If volume_type = PostgresVolumeType.HOST_PATH
    volume_host_path: Optional[str] = None

    # Use Aws Ebs as Volume
    # EbsVolume if volume_type = PostgresVolumeType.AWS_EBS
    ebs_volume: Optional[Any] = None
    # EbsVolume region is used to determine the ebs_volume_id
    # and add topology region selectors
    ebs_volume_region: Optional[str] = None
    # Provide Ebs Volume-id manually
    ebs_volume_id: Optional[str] = None
    # Add topology az selectors
    ebs_volume_az: Optional[str] = None
    # Add NodeSelectors to Pods, so they are scheduled in the same
    # region and zone as the ebs_volume
    schedule_pods_in_ebs_topology: bool = True

    # K8s Service Configuration
    create_service: bool = True
    # The port exposed by the service.
    service_port: int = 5432


class PostgresDb(DbApp):
    def __init__(
        self,
        name: str = "postgres",
        version: str = "1",
        enabled: bool = True,
        # -*- Image Configuration,
        image_name: str = "postgres",
        image_tag: str = "14",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # -*- Postgres Configuration,
        # Provide POSTGRES_USER as db_user or POSTGRES_USER in secrets_file,
        db_user: Optional[str] = None,
        # Provide POSTGRES_PASSWORD as db_password or POSTGRES_PASSWORD in secrets_file,
        db_password: Optional[str] = None,
        # Provide POSTGRES_DB as db_schema or POSTGRES_DB in secrets_file,
        db_schema: Optional[str] = None,
        db_driver: str = "postgresql",
        pgdata: Optional[str] = "/var/lib/postgresql/data/pgdata",
        postgres_initdb_args: Optional[str] = None,
        postgres_initdb_waldir: Optional[str] = None,
        postgres_host_auth_method: Optional[str] = None,
        postgres_password_file: Optional[str] = None,
        postgres_user_file: Optional[str] = None,
        postgres_db_file: Optional[str] = None,
        postgres_initdb_args_file: Optional[str] = None,
        # -*- Container Configuration,
        container_name: Optional[str] = None,
        # Add labels to the container,
        container_labels: Optional[Dict[str, Any]] = None,
        # Container env passed to the PhidataApp,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Container secrets,
        # Add secret variables to container env,
        secrets: Optional[Dict[str, str]] = None,
        # Read secret variables from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Read secret variables from AWS Secrets,
        aws_secrets: Optional[Any] = None,
        # Container ports,
        # Open a container port if open_container_port=True,
        open_container_port: bool = True,
        # Port number on the container,
        container_port: int = 5432,
        # Port name: Only used by the K8sContainer,
        container_port_name: str = "pg",
        # Host port: Only used by the DockerContainer,
        container_host_port: int = 5432,
        # Container volumes,
        create_volume: bool = True,
        volume_name: Optional[str] = None,
        volume_type: PostgresVolumeType = PostgresVolumeType.EmptyDir,
        # Container path to mount the postgres volume,
        # should be the parent directory of pgdata,
        volume_container_path: str = "/var/lib/postgresql/data",
        # Host path to mount the postgres volume,
        # If volume_type = PostgresVolumeType.HostPath,
        volume_host_path: Optional[str] = None,
        # Use Aws Ebs as Volume,
        # EbsVolume if volume_type = PostgresVolumeType.AwsEbs,
        ebs_volume: Optional[Any] = None,
        # EbsVolume region is used to determine the ebs_volume_id,
        # and add topology region selectors,
        ebs_volume_region: Optional[str] = None,
        # Provide Ebs Volume-id manually,
        ebs_volume_id: Optional[str] = None,
        # Add topology az selectors,
        ebs_volume_az: Optional[str] = None,
        # Add NodeSelectors to Pods, so they are scheduled in the same,
        # region and zone as the ebs_volume,
        schedule_pods_in_ebs_topology: bool = True,
        # -*- Docker configuration,
        # Run container in the background and return a Container object.,
        container_detach: bool = True,
        # Enable auto-removal of the container on daemon side when the container’s process exits.,
        container_auto_remove: bool = True,
        # Remove the container when it has finished running. Default: True.,
        container_remove: bool = True,
        # Username or UID to run commands as inside the container.,
        container_user: Optional[Union[str, int]] = None,
        # Keep STDIN open even if not attached.,
        container_stdin_open: bool = True,
        container_tty: bool = True,
        # Specify a test to perform to check that the container is healthy.,
        container_healthcheck: Optional[Dict[str, Any]] = None,
        # Optional hostname for the container.,
        container_hostname: Optional[str] = None,
        # Platform in the format os[/arch[/variant]].,
        container_platform: Optional[str] = None,
        # Path to the working directory.,
        container_working_dir: Optional[str] = None,
        # Restart the container when it exits. Configured as a dictionary with keys:,
        # Name: One of on-failure, or always.,
        # MaximumRetryCount: Number of times to restart the container on failure.,
        # For example: {"Name": "on-failure", "MaximumRetryCount": 5},
        container_restart_policy_docker: Optional[Dict[str, Any]] = None,
        # Add volumes to DockerContainer,
        # container_volumes is a dictionary which adds the volumes to mount,
        # inside the container. The key is either the host path or a volume name,,
        # and the value is a dictionary with 2 keys:,
        #   bind - The path to mount the volume inside the container,
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.,
        # For example:,
        # {,
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},,
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'},
        # },
        container_volumes_docker: Optional[Dict[str, dict]] = None,
        # Add ports to DockerContainer,
        # The keys of the dictionary are the ports to bind inside the container,,
        # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.,
        # The values of the dictionary are the corresponding ports to open on the host, which can be either:,
        #   - The port number, as an integer.,
        #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.,
        #   - None, to assign a random host port. For example, {'2222/tcp': None}.,
        #   - A tuple of (address, port) if you want to specify the host interface.,
        #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.,
        #   - A list of integers, if you want to bind multiple host ports to a single container port.,
        #       For example, {'1111/tcp': [1234, 4567]}.,
        container_ports_docker: Optional[Dict[str, Any]] = None,
        # -*- K8s configuration
        # K8s Deployment configuration
        replicas: int = 1,
        pod_name: Optional[str] = None,
        deploy_name: Optional[str] = None,
        secret_name: Optional[str] = None,
        configmap_name: Optional[str] = None,
        # Type: ImagePullPolicy,
        image_pull_policy: Optional[ImagePullPolicy] = None,
        pod_annotations: Optional[Dict[str, str]] = None,
        pod_node_selector: Optional[Dict[str, str]] = None,
        # Type: RestartPolicy,
        deploy_restart_policy: Optional[Any] = None,
        deploy_labels: Optional[Dict[str, Any]] = None,
        termination_grace_period_seconds: Optional[int] = None,
        # How to spread the deployment across a topology,
        # Key to spread the pods across,
        topology_spread_key: Optional[str] = None,
        # The degree to which pods may be unevenly distributed,
        topology_spread_max_skew: Optional[int] = None,
        # How to deal with a pod if it doesn't satisfy the spread constraint.,
        topology_spread_when_unsatisfiable: Optional[str] = None,
        # K8s Service Configuration,
        create_service: bool = True,
        service_name: Optional[str] = None,
        # Type: ServiceType,
        service_type: Optional[ServiceType] = None,
        # The port exposed by the service.,
        service_port: int = 5432,
        # The node_port exposed by the service if service_type = ServiceType.NODE_PORT,
        service_node_port: Optional[int] = None,
        # The target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        service_target_port: Optional[Union[str, int]] = None,
        # Extra ports exposed by the webserver service. Type: List[CreatePort],
        service_ports: Optional[List[Any]] = None,
        # Service labels,
        service_labels: Optional[Dict[str, Any]] = None,
        # Service annotations,
        service_annotations: Optional[Dict[str, str]] = None,
        # If ServiceType == ServiceType.LoadBalancer,
        service_health_check_node_port: Optional[int] = None,
        service_internal_traffic_policy: Optional[str] = None,
        service_load_balancer_class: Optional[str] = None,
        service_load_balancer_ip: Optional[str] = None,
        service_load_balancer_source_ranges: Optional[List[str]] = None,
        service_allocate_load_balancer_node_ports: Optional[bool] = None,
        # K8s RBAC Configuration,
        use_rbac: bool = False,
        # Create a Namespace with name ns_name & default values,
        ns_name: Optional[str] = None,
        # or Provide the full Namespace definition,
        # Type: CreateNamespace,
        namespace: Optional[Any] = None,
        # Create a ServiceAccount with name sa_name & default values,
        sa_name: Optional[str] = None,
        # or Provide the full ServiceAccount definition,
        # Type: CreateServiceAccount,
        service_account: Optional[Any] = None,
        # Create a ClusterRole with name cr_name & default values,
        cr_name: Optional[str] = None,
        # or Provide the full ClusterRole definition,
        # Type: CreateClusterRole,
        cluster_role: Optional[Any] = None,
        # Create a ClusterRoleBinding with name crb_name & default values,
        crb_name: Optional[str] = None,
        # or Provide the full ClusterRoleBinding definition,
        # Type: CreateClusterRoleBinding,
        cluster_role_binding: Optional[Any] = None,
        # Add additional Kubernetes resources to the App,
        # Type: CreateSecret,
        extra_secrets: Optional[List[Any]] = None,
        # Type: CreateConfigMap,
        extra_configmaps: Optional[List[Any]] = None,
        # Type: CreateService,
        extra_services: Optional[List[Any]] = None,
        # Type: CreateDeployment,
        extra_deployments: Optional[List[Any]] = None,
        # Type: CreatePersistentVolume,
        extra_pvs: Optional[List[Any]] = None,
        # Type: CreatePVC,
        extra_pvcs: Optional[List[Any]] = None,
        # Type: CreateContainer,
        extra_containers: Optional[List[Any]] = None,
        # Type: CreateContainer,
        extra_init_containers: Optional[List[Any]] = None,
        # Type: CreatePort,
        extra_ports: Optional[List[Any]] = None,
        # Type: CreateVolume,
        extra_volumes: Optional[List[Any]] = None,
        # Type: CreateStorageClass,
        extra_storage_classes: Optional[List[Any]] = None,
        # Type: CreateCustomObject,
        extra_custom_objects: Optional[List[Any]] = None,
        # Type: CreateCustomResourceDefinition,
        extra_crds: Optional[List[Any]] = None,
        # Other args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        **kwargs,
    ):
        super().__init__()
        try:
            self.args: PostgresDbArgs = PostgresDbArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                db_user=db_user,
                db_password=db_password,
                db_schema=db_schema,
                db_driver=db_driver,
                pgdata=pgdata,
                postgres_initdb_args=postgres_initdb_args,
                postgres_initdb_waldir=postgres_initdb_waldir,
                postgres_host_auth_method=postgres_host_auth_method,
                postgres_password_file=postgres_password_file,
                postgres_user_file=postgres_user_file,
                postgres_db_file=postgres_db_file,
                postgres_initdb_args_file=postgres_initdb_args_file,
                container_name=container_name,
                container_labels=container_labels,
                env=env,
                env_file=env_file,
                secrets=secrets,
                secrets_file=secrets_file,
                aws_secrets=aws_secrets,
                open_container_port=open_container_port,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                create_volume=create_volume,
                volume_name=volume_name,
                volume_type=volume_type,
                volume_container_path=volume_container_path,
                volume_host_path=volume_host_path,
                ebs_volume=ebs_volume,
                ebs_volume_region=ebs_volume_region,
                ebs_volume_id=ebs_volume_id,
                ebs_volume_az=ebs_volume_az,
                schedule_pods_in_ebs_topology=schedule_pods_in_ebs_topology,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                container_user=container_user,
                container_stdin_open=container_stdin_open,
                container_tty=container_tty,
                container_healthcheck=container_healthcheck,
                container_hostname=container_hostname,
                container_platform=container_platform,
                container_working_dir=container_working_dir,
                container_restart_policy_docker=container_restart_policy_docker,
                container_volumes_docker=container_volumes_docker,
                container_ports_docker=container_ports_docker,
                replicas=replicas,
                pod_name=pod_name,
                deploy_name=deploy_name,
                secret_name=secret_name,
                configmap_name=configmap_name,
                image_pull_policy=image_pull_policy,
                pod_annotations=pod_annotations,
                pod_node_selector=pod_node_selector,
                deploy_restart_policy=deploy_restart_policy,
                deploy_labels=deploy_labels,
                termination_grace_period_seconds=termination_grace_period_seconds,
                topology_spread_key=topology_spread_key,
                topology_spread_max_skew=topology_spread_max_skew,
                topology_spread_when_unsatisfiable=topology_spread_when_unsatisfiable,
                create_service=create_service,
                service_name=service_name,
                service_type=service_type,
                service_port=service_port,
                service_node_port=service_node_port,
                service_target_port=service_target_port,
                service_ports=service_ports,
                service_labels=service_labels,
                service_annotations=service_annotations,
                service_health_check_node_port=service_health_check_node_port,
                service_internal_traffic_policy=service_internal_traffic_policy,
                service_load_balancer_class=service_load_balancer_class,
                service_load_balancer_ip=service_load_balancer_ip,
                service_load_balancer_source_ranges=service_load_balancer_source_ranges,
                service_allocate_load_balancer_node_ports=service_allocate_load_balancer_node_ports,
                use_rbac=use_rbac,
                ns_name=ns_name,
                namespace=namespace,
                sa_name=sa_name,
                service_account=service_account,
                cr_name=cr_name,
                cluster_role=cluster_role,
                crb_name=crb_name,
                cluster_role_binding=cluster_role_binding,
                extra_secrets=extra_secrets,
                extra_configmaps=extra_configmaps,
                extra_services=extra_services,
                extra_deployments=extra_deployments,
                extra_pvs=extra_pvs,
                extra_pvcs=extra_pvcs,
                extra_containers=extra_containers,
                extra_init_containers=extra_init_containers,
                extra_ports=extra_ports,
                extra_volumes=extra_volumes,
                extra_storage_classes=extra_storage_classes,
                extra_custom_objects=extra_custom_objects,
                extra_crds=extra_crds,
                use_cache=use_cache,
                extra_kwargs=kwargs,
            )
        except Exception:
            logger.error(f"Args for {self.name} are not valid")
            raise

    def get_db_user(self) -> Optional[str]:
        db_user_var: Optional[str] = self.args.db_user if self.args else None
        if db_user_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_USER from secrets_file")
            secret_data_from_file = self.get_secret_data()
            if secret_data_from_file is not None:
                db_user_var = secret_data_from_file.get("POSTGRES_USER", db_user_var)
        return db_user_var

    def get_db_password(self) -> Optional[str]:
        db_password_var: Optional[str] = self.args.db_password if self.args else None
        if db_password_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_PASSWORD from secrets_file")
            secret_data_from_file = self.get_secret_data()
            if secret_data_from_file is not None:
                db_password_var = secret_data_from_file.get(
                    "POSTGRES_PASSWORD", db_password_var
                )
        return db_password_var

    def get_db_schema(self) -> Optional[str]:
        db_schema_var: Optional[str] = self.args.db_schema if self.args else None
        if db_schema_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            # logger.debug(f"Reading POSTGRES_DB from secrets_file")
            secret_data_from_file = self.get_secret_data()
            if secret_data_from_file is not None:
                db_schema_var = secret_data_from_file.get("POSTGRES_DB", db_schema_var)
        return db_schema_var

    def get_db_driver(self) -> Optional[str]:
        return self.args.db_driver if self.args else "postgresql"

    def get_db_host_local(self) -> Optional[str]:
        return "localhost"

    def get_db_port_local(self) -> Optional[int]:
        return self.args.container_host_port if self.args else None

    def get_db_host_docker(self) -> Optional[str]:
        return self.get_container_name()

    def get_db_port_docker(self) -> Optional[int]:
        return self.args.container_port if self.args else None

    def get_db_host_k8s(self) -> Optional[str]:
        return self.get_service_name()

    def get_db_port_k8s(self) -> Optional[int]:
        return self.get_service_port()

    def get_db_connection_url_local(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_docker(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_docker()
        port = self.get_db_port_docker()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    def get_db_connection_url_k8s(self) -> Optional[str]:
        user = self.get_db_user()
        password = self.get_db_password()
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_k8s()
        port = self.get_db_port_k8s()
        return f"{driver}://{user}:{password}@{host}:{port}/{schema}"

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(self, docker_build_context: Any) -> Optional[Any]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        from phidata.infra.docker.resource.group import (
            DockerNetwork,
            DockerContainer,
            DockerResourceGroup,
            DockerBuildContext,
        )
        from phidata.utils.common import get_default_volume_name

        if docker_build_context is None or not isinstance(
            docker_build_context, DockerBuildContext
        ):
            logger.error("docker_build_context must be a DockerBuildContext")
            return None

        # Container Environment
        container_env: Dict[str, str] = {}

        # Set postgres env vars
        # Check: https://hub.docker.com/_/postgres
        db_user = self.get_db_user()
        if db_user:
            container_env["POSTGRES_USER"] = db_user
        db_password = self.get_db_password()
        if db_password:
            container_env["POSTGRES_PASSWORD"] = db_password
        db_schema = self.get_db_schema()
        if db_schema:
            container_env["POSTGRES_DB"] = db_schema
        if self.args.pgdata:
            container_env["PGDATA"] = self.args.pgdata
        if self.args.postgres_initdb_args:
            container_env["POSTGRES_INITDB_ARGS"] = self.args.postgres_initdb_args
        if self.args.postgres_initdb_waldir:
            container_env["POSTGRES_INITDB_WALDIR"] = self.args.postgres_initdb_waldir
        if self.args.postgres_host_auth_method:
            container_env[
                "POSTGRES_HOST_AUTH_METHOD"
            ] = self.args.postgres_host_auth_method
        if self.args.postgres_password_file:
            container_env["POSTGRES_PASSWORD_FILE"] = self.args.postgres_password_file
        if self.args.postgres_user_file:
            container_env["POSTGRES_USER_FILE"] = self.args.postgres_user_file
        if self.args.postgres_db_file:
            container_env["POSTGRES_DB_FILE"] = self.args.postgres_db_file
        if self.args.postgres_initdb_args_file:
            container_env[
                "POSTGRES_INITDB_ARGS_FILE"
            ] = self.args.postgres_initdb_args_file

        # Set airflow env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        container_volumes: Dict[str, dict] = self.args.container_volumes_docker or {}
        # Create a volume for the postgres data
        if self.args.create_volume:
            if self.args.volume_type == PostgresVolumeType.EmptyDir:
                volume_name = self.args.volume_name or get_default_volume_name(app_name)
                logger.debug(f"Mounting: {volume_name}")
                logger.debug(f"\tto: {self.args.volume_container_path}")
                container_volumes[volume_name] = {
                    "bind": self.args.volume_container_path,
                    "mode": "rw",
                }
            elif self.args.volume_type == PostgresVolumeType.HostPath:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    logger.debug(f"Mounting: {volume_host_path_str}")
                    logger.debug(f"\tto: {self.args.volume_container_path}")
                    container_volumes[volume_host_path_str] = {
                        "bind": self.args.volume_container_path,
                        "mode": "rw",
                    }
                else:
                    logger.error("PostgresDb: volume_host_path not provided")
                    return None
            else:
                logger.error(f"{self.args.volume_type.value} not supported")
                return None

        # Container Ports
        container_ports: Dict[str, int] = self.args.container_ports_docker or {}

        # if open_container_port = True
        if self.args.open_container_port:
            container_ports[
                str(self.args.container_port)
            ] = self.args.container_host_port

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=self.get_image_str(),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            healthcheck=self.args.container_healthcheck,
            hostname=self.args.container_hostname,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            platform=self.args.container_platform,
            ports=container_ports if len(container_ports) > 0 else None,
            remove=self.args.container_remove,
            restart_policy=self.args.container_restart_policy_docker,
            stdin_open=self.args.container_stdin_open,
            tty=self.args.container_tty,
            user=self.args.container_user,
            volumes=container_volumes if len(container_volumes) > 0 else None,
            working_dir=self.args.container_working_dir,
            use_cache=self.args.use_cache,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(self, docker_build_context: Any) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            from collections import OrderedDict

            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(self, k8s_build_context: Any) -> Optional[Any]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        from phidata.infra.k8s.create.common.port import CreatePort
        from phidata.infra.k8s.create.core.v1.container import CreateContainer
        from phidata.infra.k8s.create.core.v1.volume import (
            CreateVolume,
            HostPathVolumeSource,
            AwsElasticBlockStoreVolumeSource,
            VolumeType,
        )
        from phidata.infra.k8s.create.group import (
            CreateK8sResourceGroup,
            CreateNamespace,
            CreateServiceAccount,
            CreateClusterRole,
            CreateClusterRoleBinding,
            CreateSecret,
            CreateConfigMap,
            CreateStorageClass,
            CreateService,
            CreateDeployment,
            CreateCustomObject,
            CreateCustomResourceDefinition,
            CreatePersistentVolume,
            CreatePVC,
        )
        from phidata.infra.k8s.resource.group import K8sBuildContext
        from phidata.utils.common import get_default_volume_name

        if k8s_build_context is None or not isinstance(
            k8s_build_context, K8sBuildContext
        ):
            logger.error("k8s_build_context must be a K8sBuildContext")
            return None

        # Init K8s resources for the CreateK8sResourceGroup
        ns: Optional[CreateNamespace] = self.args.namespace
        sa: Optional[CreateServiceAccount] = self.args.service_account
        cr: Optional[CreateClusterRole] = self.args.cluster_role
        crb: Optional[CreateClusterRoleBinding] = self.args.cluster_role_binding
        secrets: List[CreateSecret] = self.args.extra_secrets or []
        config_maps: List[CreateConfigMap] = self.args.extra_configmaps or []
        services: List[CreateService] = self.args.extra_services or []
        deployments: List[CreateDeployment] = self.args.extra_deployments or []
        pvs: List[CreatePersistentVolume] = self.args.extra_pvs or []
        pvcs: List[CreatePVC] = self.args.extra_pvcs or []
        containers: List[CreateContainer] = self.args.extra_containers or []
        init_containers: List[CreateContainer] = self.args.extra_init_containers or []
        ports: List[CreatePort] = self.args.extra_ports or []
        volumes: List[CreateVolume] = self.args.extra_volumes or []
        storage_classes: List[CreateStorageClass] = (
            self.args.extra_storage_classes or []
        )
        custom_objects: List[CreateCustomObject] = self.args.extra_custom_objects or []
        crds: List[CreateCustomResourceDefinition] = self.args.extra_crds or []

        # Common variables used by all resources
        # Use the Namespace provided with the App or
        # use the default Namespace from the k8s_build_context
        ns_name: str = self.args.ns_name or k8s_build_context.namespace
        sa_name: Optional[str] = (
            self.args.sa_name or k8s_build_context.service_account_name
        )
        common_labels: Optional[Dict[str, str]] = k8s_build_context.labels

        # -*- Use K8s RBAC
        # If use_rbac is True, use separate RBAC for this App
        # Create a namespace, service account, cluster role and cluster role binding
        if self.args.use_rbac:
            # Create Namespace for this App
            if ns is None:
                ns = CreateNamespace(
                    ns=ns_name,
                    app_name=app_name,
                    labels=common_labels,
                )
            ns_name = ns.ns

            # Create Service Account for this App
            if sa is None:
                sa = CreateServiceAccount(
                    sa_name=sa_name or self.get_sa_name(),
                    app_name=app_name,
                    namespace=ns_name,
                )
            sa_name = sa.sa_name

            # Create Cluster Role for this App
            from phidata.infra.k8s.create.rbac_authorization_k8s_io.v1.cluster_role import (
                PolicyRule,
            )

            if cr is None:
                cr = CreateClusterRole(
                    cr_name=self.args.cr_name or self.get_cr_name(),
                    rules=[
                        PolicyRule(
                            api_groups=[""],
                            resources=[
                                "pods",
                                "secrets",
                                "configmaps",
                            ],
                            verbs=[
                                "get",
                                "list",
                                "watch",
                                "create",
                                "update",
                                "patch",
                                "delete",
                            ],
                        ),
                    ],
                    app_name=app_name,
                    labels=common_labels,
                )

            # Create ClusterRoleBinding for this App
            if crb is None:
                crb = CreateClusterRoleBinding(
                    crb_name=self.args.crb_name or self.get_crb_name(),
                    cr_name=cr.cr_name,
                    service_account_name=sa.sa_name,
                    app_name=app_name,
                    namespace=ns_name,
                    labels=common_labels,
                )

        # Container Environment
        container_env: Dict[str, str] = {}

        # Set postgres env vars
        # Check: https://hub.docker.com/_/postgres
        db_user = self.get_db_user()
        if db_user:
            container_env["POSTGRES_USER"] = db_user
        db_password = self.get_db_password()
        if db_password:
            container_env["POSTGRES_PASSWORD"] = db_password
        db_schema = self.get_db_schema()
        if db_schema:
            container_env["POSTGRES_DB"] = db_schema
        if self.args.pgdata:
            container_env["PGDATA"] = self.args.pgdata
        if self.args.postgres_initdb_args:
            container_env["POSTGRES_INITDB_ARGS"] = self.args.postgres_initdb_args
        if self.args.postgres_initdb_waldir:
            container_env["POSTGRES_INITDB_WALDIR"] = self.args.postgres_initdb_waldir
        if self.args.postgres_host_auth_method:
            container_env[
                "POSTGRES_HOST_AUTH_METHOD"
            ] = self.args.postgres_host_auth_method
        if self.args.postgres_password_file:
            container_env["POSTGRES_PASSWORD_FILE"] = self.args.postgres_password_file
        if self.args.postgres_user_file:
            container_env["POSTGRES_USER_FILE"] = self.args.postgres_user_file
        if self.args.postgres_db_file:
            container_env["POSTGRES_DB_FILE"] = self.args.postgres_db_file
        if self.args.postgres_initdb_args_file:
            container_env[
                "POSTGRES_INITDB_ARGS_FILE"
            ] = self.args.postgres_initdb_args_file

        # Set airflow env vars
        self.set_aws_env_vars(env_dict=container_env)

        # Update the container env using env_file
        env_data_from_file = self.get_env_data()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.configmap_name or self.get_configmap_name(),
            app_name=app_name,
            namespace=ns_name,
            data=container_env,
            labels=common_labels,
        )
        config_maps.append(container_env_cm)

        # Create a Secret to set the container env variables which are Secret
        _secret_data = self.get_secret_data()
        if _secret_data is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or self.get_secret_name(),
                app_name=app_name,
                string_data=_secret_data,
                namespace=ns_name,
                labels=common_labels,
            )
            secrets.append(container_env_secret)

        # Container Volumes
        # Add NodeSelectors to Pods in case we create az sensitive volumes
        pod_node_selector: Optional[Dict[str, str]] = self.args.pod_node_selector
        if self.args.create_volume:
            volume_name = self.args.volume_name or get_default_volume_name(app_name)
            if self.args.volume_type == PostgresVolumeType.EmptyDir:
                pg_volume = CreateVolume(
                    volume_name=volume_name,
                    app_name=app_name,
                    mount_path=self.args.volume_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(pg_volume)
            elif self.args.volume_type == PostgresVolumeType.HostPath:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    pg_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.HOST_PATH,
                        host_path=HostPathVolumeSource(
                            path=volume_host_path_str,
                        ),
                    )
                    volumes.append(pg_volume)
                else:
                    logger.error("PostgresDb: volume_host_path not provided")
                    return None
            elif self.args.volume_type in (
                PostgresVolumeType.AwsEbs,
                PostgresVolumeType.AWS_EBS,
            ):
                if (
                    self.args.ebs_volume_id is not None
                    or self.args.ebs_volume is not None
                ):
                    # To use an EbsVolume as the volume_type we:
                    # 1. Need the volume_id
                    # 2. Need to make sure pods are scheduled in the
                    #       same region/az as the volume

                    # For the volume_id we can either:
                    # 1. Use self.args.ebs_volume_id
                    # 2. Derive it from the self.args.ebs_volume
                    ebs_volume_id = self.args.ebs_volume_id

                    # For the aws_region/az:
                    # 1. Use self.args.ebs_volume_region
                    # 2. Derive it from self.args.ebs_volume
                    # 3. Derive it from the PhidataAppArgs
                    ebs_volume_region = self.args.ebs_volume_region
                    ebs_volume_az = self.args.ebs_volume_az

                    # Derive the aws_region from self.args.ebs_volume if needed
                    if ebs_volume_region is None and self.args.ebs_volume is not None:
                        # Note: this will use the `$AWS_REGION` env var if set
                        _aws_region_from_ebs_volume = (
                            self.args.ebs_volume.get_aws_region()
                        )
                        if _aws_region_from_ebs_volume is not None:
                            ebs_volume_region = _aws_region_from_ebs_volume

                    # Derive the aws_region from the PhidataAppArgs if needed
                    if ebs_volume_region is None:
                        ebs_volume_region = self.aws_region

                    # Derive the availability_zone from self.args.ebs_volume if needed
                    if ebs_volume_az is None and self.args.ebs_volume is not None:
                        ebs_volume_az = self.args.ebs_volume.availability_zone

                    logger.debug(f"ebs_volume_region: {ebs_volume_region}")
                    logger.debug(f"ebs_volume_az: {ebs_volume_az}")

                    # Derive ebs_volume_id from self.args.ebs_volume if needed
                    if ebs_volume_id is None and self.args.ebs_volume is not None:
                        ebs_volume_id = self.args.ebs_volume.get_volume_id(
                            aws_region=ebs_volume_region,
                            aws_profile=self.aws_profile,
                        )

                    logger.debug(f"ebs_volume_id: {ebs_volume_id}")
                    if ebs_volume_id is None:
                        logger.error("Could not find volume_id for EbsVolume")
                        return None

                    pg_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.AWS_EBS,
                        aws_ebs=AwsElasticBlockStoreVolumeSource(
                            volume_id=ebs_volume_id,
                        ),
                    )
                    volumes.append(pg_volume)

                    # VERY IMPORTANT: pods should be scheduled in the same region/az as the volume
                    # To do this, we add NodeSelectors to Pods
                    if self.args.schedule_pods_in_ebs_topology:
                        if pod_node_selector is None:
                            pod_node_selector = {}

                        # Add NodeSelectors to Pods, so they are scheduled in the same
                        # region and zone as the ebs_volume
                        # https://kubernetes.io/docs/reference/labels-annotations-taints/#topologykubernetesiozone
                        if ebs_volume_region is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/region"
                            ] = ebs_volume_region

                        if ebs_volume_az is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/zone"
                            ] = ebs_volume_az
                else:
                    logger.error("PostgresDb: ebs_volume not provided")
                    return None
            else:
                logger.error(f"{self.args.volume_type.value} not supported")
                return None

        # Create the ports to open
        if self.args.open_container_port:
            container_port = CreatePort(
                name=self.args.container_port_name,
                container_port=self.args.container_port,
                service_port=self.args.service_port,
                target_port=self.args.service_target_port
                or self.args.container_port_name,
            )
            ports.append(container_port)

            # Validate NODE_PORT before adding it to the container_port
            # If ServiceType == NODE_PORT then validate self.args.service_node_port is available
            if self.args.service_type == ServiceType.NODE_PORT:
                if (
                    self.args.service_node_port is None
                    or self.args.service_node_port < 30000
                    or self.args.service_node_port > 32767
                ):
                    logger.error(f"NodePort: {self.args.service_node_port} invalid")
                    logger.error(f"Skipping this service")
                    return None
                else:
                    container_port.node_port = self.args.service_node_port
            # If ServiceType == LOAD_BALANCER then validate self.args.service_node_port only IF available
            elif self.args.service_type == ServiceType.LOAD_BALANCER:
                if self.args.service_node_port is not None:
                    if (
                        self.args.service_node_port < 30000
                        or self.args.service_node_port > 32767
                    ):
                        logger.error(f"NodePort: {self.args.service_node_port} invalid")
                        logger.error(f"Skipping this service")
                        return None
                    else:
                        container_port.node_port = self.args.service_node_port
            # else validate self.args.service_node_port is NOT available
            elif self.args.service_node_port is not None:
                logger.warning(
                    f"NodePort: {self.args.service_node_port} provided without specifying ServiceType as NODE_PORT or LOAD_BALANCER"
                )
                logger.warning("NodePort value will be ignored")
                self.args.service_node_port = None

        container_labels: Dict[str, Any] = common_labels or {}
        if self.args.container_labels is not None and isinstance(
            self.args.container_labels, dict
        ):
            container_labels.update(self.args.container_labels)

        # Create the Postgres container
        pg_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=[self.args.entrypoint]
            if isinstance(self.args.entrypoint, str)
            else self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy
            or ImagePullPolicy.IF_NOT_PRESENT,
            envs_from_configmap=[cm.cm_name for cm in config_maps]
            if len(config_maps) > 0
            else None,
            envs_from_secret=[secret.secret_name for secret in secrets]
            if len(secrets) > 0
            else None,
            ports=ports if len(ports) > 0 else None,
            volumes=volumes if len(volumes) > 0 else None,
            labels=container_labels,
        )
        # Add Postgres container to the front of the containers list
        containers.insert(0, pg_container)

        # Set default container for kubectl commands
        # https://kubernetes.io/docs/reference/labels-annotations-taints/#kubectl-kubernetes-io-default-container
        pod_annotations = {
            "kubectl.kubernetes.io/default-container": pg_container.container_name,
        }
        if self.args.pod_annotations is not None and isinstance(
            self.args.pod_annotations, dict
        ):
            pod_annotations.update(self.args.pod_annotations)

        deploy_labels: Dict[str, Any] = common_labels or {}
        if self.args.deploy_labels is not None and isinstance(
            self.args.deploy_labels, dict
        ):
            deploy_labels.update(self.args.deploy_labels)

        # Create the Postgres deployment
        pg_deployment = CreateDeployment(
            deploy_name=self.get_deploy_name(),
            pod_name=self.get_pod_name(),
            app_name=app_name,
            namespace=ns_name,
            service_account_name=sa_name,
            replicas=self.args.replicas,
            containers=containers,
            init_containers=init_containers if len(init_containers) > 0 else None,
            pod_node_selector=pod_node_selector,
            restart_policy=self.args.deploy_restart_policy or RestartPolicy.ALWAYS,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=volumes if len(volumes) > 0 else None,
            labels=deploy_labels,
            pod_annotations=pod_annotations,
            topology_spread_key=self.args.topology_spread_key,
            topology_spread_max_skew=self.args.topology_spread_max_skew,
            topology_spread_when_unsatisfiable=self.args.topology_spread_when_unsatisfiable,
        )
        deployments.append(pg_deployment)

        # Create the Postgres service
        if self.args.create_service:
            service_labels: Dict[str, Any] = common_labels or {}
            if self.args.service_labels is not None and isinstance(
                self.args.service_labels, dict
            ):
                service_labels.update(self.args.service_labels)

            pg_service = CreateService(
                service_name=self.get_service_name(),
                app_name=app_name,
                namespace=ns_name,
                service_account_name=sa_name,
                service_type=self.args.service_type,
                deployment=pg_deployment,
                ports=ports if len(ports) > 0 else None,
                labels=service_labels,
            )
            services.append(pg_service)

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            ns=ns,
            sa=sa,
            cr=cr,
            crb=crb,
            secrets=secrets if len(secrets) > 0 else None,
            config_maps=config_maps if len(config_maps) > 0 else None,
            storage_classes=storage_classes if len(storage_classes) > 0 else None,
            services=services if len(services) > 0 else None,
            deployments=deployments if len(deployments) > 0 else None,
            custom_objects=custom_objects if len(custom_objects) > 0 else None,
            crds=crds if len(crds) > 0 else None,
            pvs=pvs if len(pvs) > 0 else None,
            pvcs=pvcs if len(pvcs) > 0 else None,
        )

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: Any) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            from collections import OrderedDict

            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from time import sleep
from typing import Optional, Any, Dict, List

from botocore.exceptions import ClientError

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class AthenaQueryExecution(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/athena.html#client
    """

    resource_type = "AthenaQueryExecution"
    service_name = "athena"

    # The SQL query statements to be executed.
    query_string: str
    # A unique case-sensitive string used to ensure the request to create the query
    # is idempotent (executes only once). If another StartQueryExecution request is received,
    # the same response is returned and another query is not created. If a parameter
    # has changed, for example, the QueryString , an error is returned.
    client_request_token: Optional[str] = None
    # The database within which the query executes.
    #     QueryExecutionContext={
    #         'Database': 'string',
    #         'Catalog': 'string'
    #     },
    query_execution_context: Optional[Dict[str, str]] = None
    # Specifies information about where and how to save the results of the query execution.
    # If the query runs in a workgroup, then workgroup's settings may override query settings.
    # This affects the query results location.
    #     ResultConfiguration={
    #         'OutputLocation': 'string',
    #         'EncryptionConfiguration': {
    #             'EncryptionOption': 'SSE_S3'|'SSE_KMS'|'CSE_KMS',
    #             'KmsKey': 'string'
    #         },
    #         'ExpectedBucketOwner': 'string',
    #         'AclConfiguration': {
    #             'S3AclOption': 'BUCKET_OWNER_FULL_CONTROL'
    #         }
    #     },
    result_configuration: Optional[Dict[str, Any]] = None
    work_group: Optional[str] = None

    query_execution_id: Optional[str] = None
    query_execution_state: Optional[str] = None
    wait_for_results: bool = True

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the AthenaQueryExecution

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        # print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}
            if self.client_request_token:
                not_null_args["ClientRequestToken"] = self.client_request_token
            if self.query_execution_context:
                not_null_args["QueryExecutionContext"] = self.query_execution_context
            if self.result_configuration:
                not_null_args["ResultConfiguration"] = self.result_configuration
            if self.work_group:
                not_null_args["WorkGroup"] = self.work_group

            # Get the service_client
            service_client = self.get_service_client(aws_client)
            print_info(f"Running AthenaQuery: \n\n{self.query_string}\n")
            query_execution_response = service_client.start_query_execution(
                QueryString=self.query_string,
                **not_null_args,
            )
            # logger.debug(f"AthenaQueryExecution: {query_execution_response}")
            # logger.debug(f"AthenaQueryExecution type: {type(query_execution_response)}")

            if query_execution_response is not None:
                self.query_execution_id = query_execution_response.get(
                    "QueryExecutionId", None
                )
                print_info(f"AthenaQueryExecution Id: {self.query_execution_id}")
                self.active_resource = query_execution_response
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be started.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the AthenaQueryExecution

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            if self.query_execution_id is None:
                logger.debug("Athena QueryExecutionId not available")
                return None

            query_execution_response = service_client.get_query_execution(
                QueryExecutionId=self.query_execution_id
            )
            # logger.debug(f"AthenaQueryExecution: {query_execution_response}")
            # logger.debug(f"AthenaQueryExecution type: {type(query_execution_response)}")

            if query_execution_response is not None:
                self.query_execution_id = query_execution_response.get(
                    "QueryExecution", {}
                ).get("QueryExecutionId", None)
                self.query_execution_state = (
                    query_execution_response.get("QueryExecution", {})
                    .get("Status", {})
                    .get("State", None)
                )
                logger.debug(f"AthenaQueryExecution Id: {self.query_execution_id}")
                logger.debug(
                    f"AthenaQueryExecution State: {self.query_execution_state}"
                )
                self.active_resource = query_execution_response
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the AthenaQueryExecution

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        self.active_resource = None
        print_info(f"AthenaQueryExecution deleted: {self.name}")
        return True

    def get_results(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Streams the results of a single query execution specified by QueryExecutionId
        from the Athena query results location in Amazon S3.

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        logger.debug(
            f"Getting results for {self.get_resource_type()}: {self.get_resource_name()}"
        )
        try:
            if self.query_execution_id is None:
                print_error("Athena QueryExecutionId not available")
                return None

            self._read(aws_client)
            while self.query_execution_state in (
                None,
                "QUEUED",
                "RUNNING",
            ):
                print_info(
                    "Query State: {}".format(
                        self.query_execution_state.lower()
                        if self.query_execution_state
                        else "waiting.."
                    )
                )
                print_info("Sleeping for 5 seconds")
                sleep(5)
                self._read(aws_client)

            # Get the service_client
            service_client = self.get_service_client(aws_client)
            query_results_response = service_client.get_query_results(
                QueryExecutionId=self.query_execution_id,
            )
            # logger.debug(f"query_results_response: {query_results_response}")
            results = query_results_response.get("ResultSet", {}).get("Rows", None)

            return results
        except Exception as e:
            print_warning("AthenaQuery results unavailable. Please try again later.")
            print_warning(e)
        return None



from phidata.check.check import Check, CheckArgs


class NonNullCheckArgs(CheckArgs):
    pass


class NonNullCheck(Check):
    def __init__(self):
        super().__init__()


class RowsMatchSourceCheckArgs(CheckArgs):
    pass


class RowsMatchSourceCheck(Check):
    def __init__(self):
        super().__init__()


class ColumnCheckArgs(CheckArgs):
    pass


class ColumnCheck(Check):
    def __init__(self):
        super().__init__()


class InputsValidArgs(CheckArgs):
    pass


class InputsValidCheck(Check):
    def __init__(self):
        super().__init__()



from typing import Optional

from phidata.asset import DataAsset, DataAssetArgs
from phidata.infra.aws.api_client import AwsApiClient

from phidata.utils.log import logger


class AwsAssetArgs(DataAssetArgs):
    # Aws variables added by WorkspaceConfig().__init__()
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None

    aws_api_client: Optional[AwsApiClient] = None


class AwsAsset(DataAsset):
    def __init__(self) -> None:
        super().__init__()
        self.args: Optional[AwsAssetArgs] = None

    @property
    def aws_region(self) -> Optional[str]:
        # aws_asset not yet initialized
        if self.args is None:
            return None

        # use cached value if available
        if self.args.aws_region:
            return self.args.aws_region

        # logger.info(f"Loading {AWS_REGION_ENV_VAR} from env")
        import os
        from phidata.constants import AWS_REGION_ENV_VAR

        aws_region_env = os.getenv(AWS_REGION_ENV_VAR)
        if aws_region_env is not None:
            self.args.aws_region = aws_region_env
        return self.args.aws_region

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        # aws_asset not yet initialized
        if self.args is None:
            return None

        # use cached value if available
        if self.args.aws_profile:
            return self.args.aws_profile

        import os
        from phidata.constants import AWS_PROFILE_ENV_VAR

        aws_profile_env = os.getenv(AWS_PROFILE_ENV_VAR)
        if aws_profile_env is not None:
            self.args.aws_profile = aws_profile_env
        return self.args.aws_profile

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_api_client(self) -> Optional[AwsApiClient]:
        # aws_asset not yet initialized
        if self.args is None:
            return None

        # use cached value if available
        if self.args.aws_api_client:
            return self.args.aws_api_client

        aws_api_client = AwsApiClient(
            aws_region=self.aws_region,
            aws_profile=self.aws_profile,
        )
        if aws_api_client is not None:
            self.args.aws_api_client = aws_api_client
        return self.args.aws_api_client

    @aws_api_client.setter
    def aws_api_client(self, aws_api_client: AwsApiClient) -> None:
        if self.args is not None and aws_api_client is not None:
            self.args.aws_api_client = aws_api_client



from collections import OrderedDict
from datetime import datetime, timedelta
from typing import Optional, Any, List, Dict, Literal

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.asset import DataAsset
from phidata.check import Check
from phidata.exceptions.workflow import WorkflowFailure
from phidata.utils.context import get_run_date, build_path_context_from_env
from phidata.utils.cli_console import print_info, print_subheading
from phidata.utils.log import logger
from phidata.types.run_status import RunStatus
from phidata.types.context import PathContext, RunContext, AirflowContext
from phidata.workflow import Workflow
from phidata.workflow.workflow_relatives import WorkflowRelatives
from phidata.utils.env_var import validate_env_vars


class DataProductArgs(PhidataBaseArgs):
    # List of workflows in this DataProduct
    workflows: Optional[List[Workflow]] = None
    # A dependency graph of the workflows in the format:
    # { downstream: [upstream_list] }
    # The downstream workflow will execute after all workflows in the
    #   upstream_list are finished.
    # upstream_list: workflows which run before the downstream workflow
    # Ex: To run extract -> transform -> load
    #   graph = {
    #     transform: [extract],
    #     load: [transform],
    #     ...
    #   }
    graph: Optional[Dict[Workflow, List[Workflow]]] = None
    # What to do when a workflow fails while running the data_product locally.
    # Continue to next task or stop?
    on_failure: Literal["stop", "continue"] = "stop"

    # Add an EmptyOperator at the start of the workflows
    add_start_task: bool = True
    # Add an EmptyOperator at the end of the workflows
    add_end_task: bool = False

    # airflow dag_id for this data_product, use name if not provided
    dag_id: Optional[str] = None
    # run_context provided by wf_operator
    run_context: Optional[RunContext] = None
    # path_context provided by env variables
    path_context: Optional[PathContext] = None
    # airflow_context provided by airflow
    airflow_context: Optional[AirflowContext] = None
    # Env vars that validate Airflow is active on the containers
    # This is used to gate code blocks which should only run on
    # remote containers like creating DAGs and tasks
    validate_airflow_env: Dict[str, Any] = {"INIT_AIRFLOW": True}

    # Input DataAssets used to build sensors before kicking off the data_product
    inputs: Optional[List[DataAsset]] = None
    # DataAssets produced by this data_product, used for building the lineage graph
    outputs: Optional[List[DataAsset]] = None

    # Checks to run before the data_product
    pre_checks: Optional[List[Check]] = None
    # Checks to run after the data_product
    post_checks: Optional[List[Check]] = None

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.validate_airflow_env)

    @property
    def run_date(self) -> datetime:
        return get_run_date(self.run_context, self.airflow_context)

    @classmethod
    def from_kwargs(cls, kwargs: Optional[Dict] = None):
        if kwargs is None or not isinstance(kwargs, dict):
            return cls()
        # logger.info(f"Loading {cls.__name__} using kwargs")
        args_object = cls(**kwargs)
        validate_airflow_env = kwargs.get("validate_airflow_env", None)
        if validate_env_vars(validate_airflow_env):
            # logger.info("Creating airflow_context")
            airflow_context = AirflowContext(**kwargs)
            args_object.airflow_context = airflow_context
        return args_object


class DataProduct(PhidataBase):
    """Base Class for all DataProducts"""

    def __init__(
        self,
        name: str,
        # List of workflows in this DataProduct,
        workflows: Optional[List[Workflow]] = None,
        # A dependency graph of the workflows in the format:
        # { downstream: [upstream_list] }
        # The downstream workflow will execute after all workflows in the
        #   upstream_list are finished.
        # upstream_list: workflows which run before the downstream workflow
        # Ex: To run extract -> transform -> load
        #   graph = {
        #     transform: [extract],
        #     load: [transform],
        #     ...
        #   }
        graph: Optional[Dict[Workflow, List[Workflow]]] = None,
        # What to do when a task fails while running the workflow locally.
        # Continue to next task or stop?
        on_failure: Literal["stop", "continue"] = "stop",
        # Add an EmptyOperator at the start of the workflows
        add_start_task: bool = True,
        # Add an EmptyOperator at the end of the workflows
        add_end_task: bool = False,
        # airflow dag_id for this workflow, use name if not provided
        dag_id: Optional[str] = None,
        # run_context provided by wf_operator
        run_context: Optional[RunContext] = None,
        # path_context provided by env variables
        path_context: Optional[PathContext] = None,
        # airflow_context provided by airflow
        airflow_context: Optional[AirflowContext] = None,
        # Input DataAssets used to build sensors before kicking off the workflow
        inputs: Optional[List[DataAsset]] = None,
        # DataAssets produced by this workflow, used for building the lineage graph
        outputs: Optional[List[DataAsset]] = None,
        # Checks to run before the workflow
        pre_checks: Optional[List[Check]] = None,
        # Checks to run after the workflow
        post_checks: Optional[List[Check]] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ) -> None:
        super().__init__()
        self._family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = None
        try:
            self.args: DataProductArgs = DataProductArgs(
                name=name,
                workflows=workflows,
                graph=graph,
                on_failure=on_failure,
                add_start_task=add_start_task,
                add_end_task=add_end_task,
                dag_id=dag_id,
                run_context=run_context,
                path_context=path_context,
                airflow_context=airflow_context,
                inputs=inputs,
                outputs=outputs,
                pre_checks=pre_checks,
                post_checks=post_checks,
                version=version,
                enabled=enabled,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def workflows(self) -> Optional[List[Workflow]]:
        return self.args.workflows if self.args else None

    @workflows.setter
    def workflows(self, workflows: List[Workflow]) -> None:
        if self.args is not None and workflows is not None:
            self.args.workflows = workflows

    @property
    def graph(self) -> Optional[Dict[Workflow, List[Workflow]]]:
        return self.args.graph if self.args else None

    @graph.setter
    def graph(self, graph: Dict[Workflow, List[Workflow]]) -> None:
        if self.args is not None and graph is not None:
            self.args.graph = graph

    @property
    def on_failure(self) -> Literal["stop", "continue"]:
        return self.args.on_failure if self.args else "stop"

    @on_failure.setter
    def on_failure(self, on_failure: Literal["stop", "continue"]) -> None:
        if self.args is not None and on_failure is not None:
            self.args.on_failure = on_failure

    @property
    def add_start_task(self) -> bool:
        return self.args.add_start_task if self.args else False

    @add_start_task.setter
    def add_start_task(self, add_start_task: bool) -> None:
        if self.args is not None and add_start_task is not None:
            self.args.add_start_task = add_start_task

    @property
    def add_end_task(self) -> bool:
        return self.args.add_end_task if self.args else False

    @add_end_task.setter
    def add_end_task(self, add_end_task: bool) -> None:
        if self.args is not None and add_end_task is not None:
            self.args.add_end_task = add_end_task

    @property
    def dag_id(self) -> str:
        return self.args.dag_id if (self.args and self.args.dag_id) else self.name

    @dag_id.setter
    def dag_id(self, dag_id: str) -> None:
        if self.args is not None and dag_id is not None:
            self.args.dag_id = dag_id

    @property
    def run_context(self) -> Optional[RunContext]:
        return self.args.run_context if self.args else None

    @run_context.setter
    def run_context(self, run_context: RunContext) -> None:
        if self.args is not None and run_context is not None:
            self.args.run_context = run_context

    @property
    def path_context(self) -> Optional[PathContext]:
        # Workflow not yet initialized
        if self.args is None:
            return None

        if self.args.path_context is not None:
            # use cached value if available
            return self.args.path_context

        self.args.path_context = build_path_context_from_env()
        return self.args.path_context

    @path_context.setter
    def path_context(self, path_context: PathContext) -> None:
        if self.args is not None and path_context is not None:
            self.args.path_context = path_context

    @property
    def airflow_context(self) -> Optional[AirflowContext]:
        return self.args.airflow_context if self.args else None

    @airflow_context.setter
    def airflow_context(self, airflow_context: AirflowContext) -> None:
        if self.args is not None and airflow_context is not None:
            self.args.airflow_context = airflow_context

    @property
    def inputs(self) -> Optional[List[DataAsset]]:
        return self.args.inputs if self.args else None

    @inputs.setter
    def inputs(self, inputs: List[DataAsset]) -> None:
        if self.args is not None and inputs is not None:
            self.args.inputs = inputs

    @property
    def outputs(self) -> Optional[List[DataAsset]]:
        return self.args.outputs if self.args else None

    @outputs.setter
    def outputs(self, outputs: List[DataAsset]) -> None:
        if self.args is not None and outputs is not None:
            self.args.outputs = outputs

    @property
    def pre_checks(self) -> Optional[List[Check]]:
        return self.args.pre_checks if self.args else None

    @pre_checks.setter
    def pre_checks(self, pre_checks: List[Check]) -> None:
        if self.args is not None and pre_checks is not None:
            self.args.pre_checks = pre_checks

    @property
    def post_checks(self) -> Optional[List[Check]]:
        return self.args.post_checks if self.args else None

    @post_checks.setter
    def post_checks(self, post_checks: List[Check]) -> None:
        if self.args is not None and post_checks is not None:
            self.args.post_checks = post_checks

    ######################################################
    ## DataProduct Properties
    ######################################################

    @property
    def airflow_active(self) -> bool:
        return validate_env_vars(self.args.validate_airflow_env)

    @property
    def family_tree(self) -> Optional[Dict[Workflow, WorkflowRelatives]]:
        if self.workflows is None:
            return None

        # used cached value if available
        if self._family_tree is not None:
            return self._family_tree

        # build family_tree
        family_tree: Dict[Workflow, WorkflowRelatives] = OrderedDict()
        for workflow in self.workflows:
            family_tree[workflow] = WorkflowRelatives()

        if self.graph is None:
            return family_tree

        for workflow, upstream_list in self.graph.items():
            family_tree[workflow].add_upstream_list(upstream_list)
            for upstream_workflow in upstream_list:
                family_tree[upstream_workflow].add_downstream(workflow)

        return family_tree

    @property
    def family_tree_dict(self) -> Dict[str, Dict[str, List[str]]]:
        tree: Dict[str, Dict[str, List[str]]] = OrderedDict()
        family_tree = self.family_tree
        if family_tree is None:
            return tree

        for workflow in family_tree:
            workflow_tree = {}
            upstream_list = family_tree[workflow].upstream_list
            downstream_list = family_tree[workflow].downstream_list
            if upstream_list:
                workflow_tree["upstream"] = [wf.name for wf in upstream_list]
            if downstream_list:
                workflow_tree["downstream"] = [wf.name for wf in downstream_list]
            tree[workflow.name] = workflow_tree
        return tree

    @property
    def roots(self) -> Optional[List[Workflow]]:
        """
        Roots are workflows that run first. i.e. have nothing upstream of them
        """
        if self.graph is None:
            return self.workflows

        family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = self.family_tree
        if family_tree is None:
            return self.workflows

        roots: List[Workflow] = []
        for workflow, relatives in family_tree.items():
            if relatives.upstream_list is None:
                roots.append(workflow)
        return roots

    @property
    def leaves(self) -> Optional[List[Workflow]]:
        """
        Leaves are workflows that run last. i.e. have nothing downstream of them
        """
        if self.graph is None:
            return self.workflows

        family_tree: Optional[Dict[Workflow, WorkflowRelatives]] = self.family_tree
        if family_tree is None:
            return self.workflows

        leaves: List[Workflow] = []
        for workflow, relatives in family_tree.items():
            if relatives.downstream_list is None:
                leaves.append(workflow)
        return leaves

    def __call__(self, *args, **kwargs) -> bool:
        logger.warning("DataProducts should not be called directly")
        return True

    ######################################################
    ## Build data_product
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Workflows
    ######################################################

    def add_workflow(self, workflow: Workflow):
        if self.args.workflows is None:
            self.args.workflows = []
        self.args.workflows.append(workflow)

    ######################################################
    ## Checks
    ######################################################

    def add_pre_check(self, check: Check):
        if self.args.pre_checks is None:
            self.args.pre_checks = []
        self.args.pre_checks.append(check)

    def add_post_check(self, check: Check):
        if self.args.post_checks is None:
            self.args.post_checks = []
        self.args.post_checks.append(check)

    ######################################################
    ## Run data_product
    ######################################################

    def run_workflow_in_local_env(
        self, workflow: Workflow, task_id: Optional[str] = None
    ) -> RunStatus:
        wf_name = workflow.name
        print_subheading(f"\nRunning workflow: {wf_name}")
        # Pass down context
        workflow.run_context = self.run_context
        workflow.path_context = self.path_context
        workflow.airflow_context = self.airflow_context
        # Not required for local runs but adding for posterity
        if self.dag_id:
            workflow.dag_id = self.dag_id
        run_success = workflow.run_in_local_env(task_id=task_id)
        return RunStatus(name=wf_name, success=run_success)

    def run_in_local_env(
        self, workflow_name: Optional[str] = None, task_id: Optional[str] = None
    ) -> bool:
        """
        Runs a DataProduct in the local environment where phi wf is called from.

        Returns:
            run_status (bool): True if the run was successful
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running DataProduct locally")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        if workflow_name is not None:
            wf_name_map = {wf.name: wf for wf in self.workflows}
            if workflow_name in wf_name_map:
                wf_run_status.append(
                    self.run_workflow_in_local_env(
                        workflow=wf_name_map[workflow_name], task_id=task_id
                    )
                )
            else:
                logger.error(f"Could not find {workflow_name} in {wf_name_map.keys()}")
        else:
            for wf in self.workflows:
                run_status = self.run_workflow_in_local_env(wf)
                wf_run_status.append(run_status)
                if not run_status.success and self.args.on_failure == "stop":
                    # logger.error(f"Workflow {wf.name} failed")
                    break

        print_run_status_table("Workflow Run Status", wf_run_status)
        for run_status in wf_run_status:
            if not run_status.success:
                return False
        return True

    def run_workflow_in_docker_container(
        self,
        workflow: Workflow,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
        task_id: Optional[str] = None,
    ) -> RunStatus:
        _wf_name = workflow.name
        print_subheading(f"\nRunning workflow: {_wf_name}")
        # Pass down context
        workflow.run_context = self.run_context
        workflow.path_context = self.path_context
        workflow.airflow_context = self.airflow_context
        workflow.dag_id = self.dag_id
        run_success = workflow.run_in_docker_container(
            active_container=active_container, docker_env=docker_env, task_id=task_id
        )
        return RunStatus(name=_wf_name, success=run_success)

    def run_in_docker_container(
        self,
        active_container: Any,
        docker_env: Optional[Dict[str, str]] = None,
        workflow_name: Optional[str] = None,
        task_id: Optional[str] = None,
    ) -> bool:
        """
        Runs a DataProduct in a docker container.

        Args:
            active_container:
            docker_env:
            workflow_name:
            task_id:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the docker container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        from phidata.utils.cli_console import print_run_status_table

        logger.info("--**-- Running DataProduct in docker container")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        if workflow_name is not None:
            wf_name_map = {wf.name: wf for wf in self.workflows}
            if workflow_name in wf_name_map:
                wf_run_status.append(
                    self.run_workflow_in_docker_container(
                        workflow=wf_name_map[workflow_name],
                        active_container=active_container,
                        docker_env=docker_env,
                        task_id=task_id,
                    )
                )
            else:
                logger.error(f"Could not find {workflow_name} in {wf_name_map.keys()}")
        else:
            for wf in self.workflows:
                run_status = self.run_workflow_in_docker_container(
                    workflow=wf,
                    active_container=active_container,
                    docker_env=docker_env,
                )
                wf_run_status.append(run_status)
                if not run_status.success and self.args.on_failure == "stop":
                    raise WorkflowFailure(f"Workflow {wf.name} failed")

        print_run_status_table("Workflow Run Status", wf_run_status)
        for run_status in wf_run_status:
            if not run_status.success:
                return False
        return True

    def run_in_k8s_container(
        self,
        pod: Any,
        k8s_api_client: Any,
        container_name: Optional[str] = None,
        k8s_env: Optional[Dict[str, str]] = None,
    ) -> bool:
        """
        Runs a DataProduct in a k8s container.

        Args:
            pod: The pod to run the workflow in
            k8s_api_client: K8sApiClient
            container_name: The container to run the workflow in
            k8s_env:

        Returns:
            run_status (bool): True if the run was successful

        Notes:
            * This function runs in the local environment where phi wf is called from.
            But executes `airflow` commands in the k8s container to run the workflow
            * For the airflow tasks to be available, they need to be added to the DAG
            using add_airflow_tasks_to_dag()
        """
        logger.info("--**-- Running DataProduct in k8s container")

        if self.workflows is None:
            logger.info("No workflows to run")
            return True

        wf_run_status: List[RunStatus] = []
        for idx, wf in enumerate(self.workflows, start=1):
            wf_name = wf.name or "{}__{}".format(wf.__class__.__name__, idx)
            print_subheading(f"\nRunning {wf_name}")
            # Pass down context
            wf.run_context = self.run_context
            wf.path_context = self.path_context
            wf.airflow_context = self.airflow_context
            wf.dag_id = self.dag_id
            run_success = wf.run_in_k8s_container(
                pod=pod,
                k8s_api_client=k8s_api_client,
                container_name=container_name,
                k8s_env=k8s_env,
            )
            wf_run_status.append(RunStatus(name=wf_name, success=run_success))

        print_subheading("\nWorkflow run status:")
        print_info(
            "\n".join(
                [
                    "{}: {}".format(wf.name, "Success" if wf.success else "Fail")
                    for wf in wf_run_status
                ]
            )
        )
        print_info("")
        for _run in wf_run_status:
            if not _run.success:
                return False
        return True

    ######################################################
    ## Airflow functions
    ######################################################

    def create_airflow_dag(
        self,
        owner: Optional[str] = "airflow",
        depends_on_past: Optional[bool] = False,
        # The description for the DAG to e.g. be shown on the webserver
        description: Optional[str] = None,
        # Defines how often that DAG runs, this
        #  timedelta object gets added to your latest task instance's
        #  execution_date to figure out the next schedule
        # Default: daily
        schedule_interval: timedelta = timedelta(days=1),
        # The timestamp from which the scheduler will
        #  attempt to backfill
        start_date: Optional[datetime] = None,
        # if start_date is not provided, use start_days_ago
        start_days_ago: int = 1,
        # A date beyond which your DAG won't run, leave to None
        #  for open ended scheduling
        end_date: Optional[datetime] = None,
        # a dictionary of macros that will be exposed
        #  in your jinja templates. For example, passing ``dict(foo='bar')``
        #  to this argument allows you to ``{{ foo }}`` in all jinja
        #  templates related to this DAG. Note that you can pass any
        #  type of object here.
        user_defined_macros: Optional[Dict] = None,
        # a dictionary of filters that will be exposed
        #  in your jinja templates. For example, passing
        #  ``dict(hello=lambda name: 'Hello %s' % name)`` to this argument allows
        #  you to ``{{ 'world' | hello }}`` in all jinja templates related to
        #  this DAG.
        user_defined_filters: Optional[Dict] = None,
        # A dictionary of default parameters to be used
        #  as constructor keyword parameters when initialising operators.
        #  Note that operators have the same hook, and precede those defined
        #  here, meaning that if your dict contains `'depends_on_past': True`
        #  here and `'depends_on_past': False` in the operator's call
        #  `default_args`, the actual value will be `False`.
        default_args: Optional[Dict] = None,
        # the number of task instances allowed to run concurrently
        concurrency: Optional[int] = None,
        # maximum number of active DAG runs, beyond this
        #  number of DAG runs in a running state, the scheduler won't create
        #  new active DAG runs
        max_active_runs: int = 8,
        doc_md: Optional[str] = None,
        # a dictionary of DAG level parameters that are made
        # accessible in templates, namespaced under `params`. These
        # params can be overridden at the task level.
        params: Optional[Dict] = None,
        is_paused_upon_creation: Optional[bool] = None,
        jinja_environment_kwargs: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ) -> Any:
        """
        Used to create an Airflow DAG for the DataProduct.
        """

        # Skips DAG creation on local machines
        if not self.airflow_active:
            return None

        from airflow import DAG
        from airflow.models.operator import Operator
        from airflow.utils.dates import days_ago

        operator_default_args = default_args
        if operator_default_args is None:
            operator_default_args = {}
        operator_default_args.update(
            {
                "owner": owner,
                "depends_on_past": depends_on_past,
            }
        )
        if start_date is None:
            start_date = days_ago(start_days_ago)

        dag_id = self.dag_id
        if dag_id is None:
            logger.error("DataProduct dag_id unavailable")
            return None
        if self.workflows is None:
            logger.error(f"DataProduct has no workflows")
            return None
        dag = DAG(
            dag_id=dag_id,
            description=description,
            schedule_interval=schedule_interval,
            start_date=start_date,
            end_date=end_date,
            user_defined_macros=user_defined_macros,
            user_defined_filters=user_defined_filters,
            default_args=operator_default_args,
            concurrency=concurrency,
            max_active_runs=max_active_runs,
            doc_md=doc_md,
            params=params,
            is_paused_upon_creation=is_paused_upon_creation,
            jinja_environment_kwargs=jinja_environment_kwargs,
            tags=tags,
        )

        # Add airflow tasks from workflows to DAG
        for workflow in self.workflows:
            # Add start tasks to workflows if there are inter-workflow dependencies
            if self.add_start_task or self.graph is not None:
                workflow.add_start_task = True
            if self.add_end_task or self.graph is not None:
                workflow.add_end_task = True
            # logger.info(f"Adding tasks for workflow: {workflow.name}")
            add_task_success = workflow.add_airflow_tasks_to_dag(dag=dag)
            if not add_task_success:
                logger.error(f"Tasks for workflow {workflow.name} could not be added")

        if self.graph is not None:
            # logger.info("Updating dependencies")
            for downstream_workflow, upstream_workflow_list in self.graph.items():

                downstream_wf_start_task: Optional[Operator] = None
                downstream_wf_start_task_id = downstream_workflow.start_workflow_task_id
                # logger.info(f"downstream_workflow_start: {downstream_wf_start_task_id}")
                if dag.has_task(downstream_wf_start_task_id):
                    downstream_wf_start_task = dag.get_task(downstream_wf_start_task_id)
                if downstream_wf_start_task is None:
                    logger.info(f"Cannot find task: {downstream_wf_start_task_id}")
                    continue

                for workflow in upstream_workflow_list:
                    upstream_wf_end_task: Optional[Operator] = None
                    upstream_wf_end_task_id = workflow.end_workflow_task_id
                    # logger.info(f"upstream_workflow_end: {upstream_wf_end_task_id}")

                    if dag.has_task(upstream_wf_end_task_id):
                        upstream_wf_end_task = dag.get_task(upstream_wf_end_task_id)

                    if upstream_wf_end_task is not None:
                        downstream_wf_start_task.set_upstream(upstream_wf_end_task)

                # logger.info(f"Upstream: {downstream_wf_start_task.upstream_task_ids}")
        logger.debug(f"Airflow dag {dag_id} created")
        return dag



from typing import Optional

from phidata.constants import PHIDATA_RUNTIME_ENV_VAR
from phidata.utils.enums import ExtendedEnum


class PhidataRuntimeType(ExtendedEnum):
    local = "local"
    docker = "docker"
    kubernetes = "kubernetes"


def get_phidata_runtime() -> Optional[PhidataRuntimeType]:
    from os import getenv

    phidata_runtime_env_var = getenv(PHIDATA_RUNTIME_ENV_VAR)
    # logger.debug(f"{PHIDATA_RUNTIME_ENV_VAR}: {phidata_runtime_env_var}")

    if (
        phidata_runtime_env_var is not None
        and phidata_runtime_env_var in PhidataRuntimeType.values_list()
    ):
        return PhidataRuntimeType.from_str(phidata_runtime_env_var)
    return PhidataRuntimeType.local



from typing import Optional, Any, Dict, List, Literal

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger


class EmrCluster(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/emr.html
    """

    resource_type = "EmrCluster"
    service_name = "emr"

    # Name of the cluster.
    name: str
    # The location in Amazon S3 to write the log files of the job flow.
    # If a value is not provided, logs are not created.
    log_uri: Optional[str] = None
    # The KMS key used for encrypting log files. If a value is not provided, the logs remain encrypted by AES-256.
    # This attribute is only available with Amazon EMR version 5.30.0 and later, excluding Amazon EMR 6.0.0.
    log_encryption_kms_key_id: Optional[str] = None
    # A JSON string for selecting additional features.
    additional_info: Optional[str] = None
    # The Amazon EMR release label, which determines the version of open-source application packages installed on the
    # cluster. Release labels are in the form emr-x.x.x,
    # where x.x.x is an Amazon EMR release version such as emr-5.14.0 .
    release_label: Optional[str] = None
    # A specification of the number and type of Amazon EC2 instances.
    instances: Optional[Dict[str, Any]] = None
    # A list of steps to run.
    steps: Optional[List[Dict[str, Any]]] = None
    # A list of bootstrap actions to run before Hadoop starts on the cluster nodes.
    bootstrap_actions: Optional[List[Dict[str, Any]]] = None
    # For Amazon EMR releases 3.x and 2.x. For Amazon EMR releases 4.x and later, use Applications.
    # A list of strings that indicates third-party software to use.
    supported_products: Optional[List[str]]
    new_supported_products: Optional[List[Dict[str, Any]]] = None
    # Applies to Amazon EMR releases 4.0 and later.
    # A case-insensitive list of applications for Amazon EMR to install and configure when launching the cluster.
    applications: Optional[List[Dict[str, Any]]] = None
    # For Amazon EMR releases 4.0 and later. The list of configurations supplied for the EMR cluster you are creating.
    configurations: Optional[List[Dict[str, Any]]] = None
    # Also called instance profile and EC2 role. An IAM role for an EMR cluster.
    # The EC2 instances of the cluster assume this role. The default role is EMR_EC2_DefaultRole.
    # In order to use the default role, you must have already created it using the CLI or console.
    job_flow_role: Optional[str] = None
    # he IAM role that Amazon EMR assumes in order to access Amazon Web Services resources on your behalf.
    service_role: Optional[str] = None
    # A list of tags to associate with a cluster and propagate to Amazon EC2 instances.
    tags: Optional[List[Dict[str, str]]] = None
    # The name of a security configuration to apply to the cluster.
    security_configuration: Optional[str] = None
    # An IAM role for automatic scaling policies. The default role is EMR_AutoScaling_DefaultRole.
    # The IAM role provides permissions that the automatic scaling feature requires to launch and terminate EC2
    # instances in an instance group.
    auto_scaling_role: Optional[str] = None
    scale_down_behavior: Optional[
        Literal["TERMINATE_AT_INSTANCE_HOUR", "TERMINATE_AT_TASK_COMPLETION"]
    ] = None
    custom_ami_id: Optional[str] = None
    # The size, in GiB, of the Amazon EBS root device volume of the Linux AMI that is used for each EC2 instance.
    ebs_root_volume_size: Optional[int] = None
    repo_upgrade_on_boot: Optional[Literal["SECURITY", "NONE"]] = None
    # Attributes for Kerberos configuration when Kerberos authentication is enabled using a security configuration.
    kerberos_attributes: Optional[Dict[str, str]] = None
    # Specifies the number of steps that can be executed concurrently.
    # The default value is 1 . The maximum value is 256 .
    step_concurrency_level: Optional[int] = None
    # The specified managed scaling policy for an Amazon EMR cluster.
    managed_scaling_policy: Optional[Dict[str, Any]] = None
    placement_group_configs: Optional[List[Dict[str, Any]]] = None
    # The auto-termination policy defines the amount of idle time in seconds after which a cluster terminates.
    auto_termination_policy: Optional[Dict[str, int]] = None

    # provided by api on create
    # A unique identifier for the job flow.
    job_flow_id: Optional[str] = None
    # The Amazon Resource Name (ARN) of the cluster.
    cluster_arn: Optional[str] = None
    # ClusterSummary returned on read
    cluster_summary: Optional[Dict] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the EmrCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # create a dict of args which are not null, otherwise aws type validation fails
            not_null_args: Dict[str, Any] = {}

            if self.log_uri:
                not_null_args["LogUri"] = self.log_uri
            if self.log_encryption_kms_key_id:
                not_null_args["LogEncryptionKmsKeyId"] = self.log_encryption_kms_key_id
            if self.additional_info:
                not_null_args["AdditionalInfo"] = self.additional_info
            if self.release_label:
                not_null_args["ReleaseLabel"] = self.release_label
            if self.instances:
                not_null_args["Instances"] = self.instances
            if self.steps:
                not_null_args["Steps"] = self.steps
            if self.bootstrap_actions:
                not_null_args["BootstrapActions"] = self.bootstrap_actions
            if self.supported_products:
                not_null_args["SupportedProducts"] = self.supported_products
            if self.new_supported_products:
                not_null_args["NewSupportedProducts"] = self.new_supported_products
            if self.applications:
                not_null_args["Applications"] = self.applications
            if self.configurations:
                not_null_args["Configurations"] = self.configurations
            if self.job_flow_role:
                not_null_args["JobFlowRole"] = self.job_flow_role
            if self.service_role:
                not_null_args["ServiceRole"] = self.service_role
            if self.tags:
                not_null_args["Tags"] = self.tags
            if self.security_configuration:
                not_null_args["SecurityConfiguration"] = self.security_configuration
            if self.auto_scaling_role:
                not_null_args["AutoScalingRole"] = self.auto_scaling_role
            if self.scale_down_behavior:
                not_null_args["ScaleDownBehavior"] = self.scale_down_behavior
            if self.custom_ami_id:
                not_null_args["CustomAmiId"] = self.custom_ami_id
            if self.ebs_root_volume_size:
                not_null_args["EbsRootVolumeSize"] = self.ebs_root_volume_size
            if self.repo_upgrade_on_boot:
                not_null_args["RepoUpgradeOnBoot"] = self.repo_upgrade_on_boot
            if self.kerberos_attributes:
                not_null_args["KerberosAttributes"] = self.kerberos_attributes
            if self.step_concurrency_level:
                not_null_args["StepConcurrencyLevel"] = self.step_concurrency_level
            if self.managed_scaling_policy:
                not_null_args["ManagedScalingPolicy"] = self.managed_scaling_policy
            if self.placement_group_configs:
                not_null_args["PlacementGroupConfigs"] = self.placement_group_configs
            if self.auto_termination_policy:
                not_null_args["AutoTerminationPolicy"] = self.auto_termination_policy

            # Get the service_client
            service_client = self.get_service_client(aws_client)

            # Create EmrCluster
            create_response = service_client.run_job_flow(
                Name=self.name,
                **not_null_args,
            )
            logger.debug(f"create_response type: {type(create_response)}")
            logger.debug(f"create_response: {create_response}")

            self.job_flow_id = create_response.get("JobFlowId", None)
            self.cluster_arn = create_response.get("ClusterArn", None)
            self.active_resource = create_response
            if self.active_resource is not None:
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} created"
                )
                logger.debug(f"JobFlowId: {self.job_flow_id}")
                logger.debug(f"ClusterArn: {self.cluster_arn}")
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def post_create(self, aws_client: AwsApiClient) -> bool:
        ## Wait for Cluster to be created
        if self.wait_for_creation:
            try:
                print_info("Waiting for EmrCluster to be active.")
                if self.job_flow_id is not None:
                    waiter = self.get_service_client(aws_client).get_waiter(
                        "cluster_running"
                    )
                    waiter.wait(
                        ClusterId=self.job_flow_id,
                        WaiterConfig={
                            "Delay": self.waiter_delay,
                            "MaxAttempts": self.waiter_max_attempts,
                        },
                    )
                else:
                    print_warning("Skipping waiter, No ClusterId found")
            except Exception as e:
                print_error("Waiter failed.")
                print_error(e)
        return True

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the EmrCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """
        from botocore.exceptions import ClientError

        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            service_client = self.get_service_client(aws_client)
            list_response = service_client.list_clusters()
            # logger.debug(f"list_response type: {type(list_response)}")
            # logger.debug(f"list_response: {list_response}")

            cluster_summary_list = list_response.get("Clusters", None)
            if cluster_summary_list is not None and isinstance(
                cluster_summary_list, list
            ):
                for _cluster_summary in cluster_summary_list:
                    cluster_name = _cluster_summary.get("Name", None)
                    if cluster_name == self.name:
                        self.active_resource = _cluster_summary
                        break

            if self.active_resource is None:
                logger.debug(f"No {self.get_resource_type()} found")
                return None

            # logger.debug(f"EmrCluster: {self.active_resource}")
            self.job_flow_id = self.active_resource.get("Id", None)
            self.cluster_arn = self.active_resource.get("ClusterArn", None)
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the EmrCluster

        Args:
            aws_client: The AwsApiClient for the current cluster
        """

        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")
        try:
            # populate self.job_flow_id
            self._read(aws_client)

            service_client = self.get_service_client(aws_client)
            self.active_resource = None

            if self.job_flow_id:
                delete_response = service_client.terminate_job_flows(
                    JobFlowIds=[self.job_flow_id]
                )
                print_info(
                    f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
                )
            else:
                print_error("Could not find cluster id")
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False



from typing import Optional, Any

from phidata.infra.base.api_client import InfraApiClient
from phidata.infra.aws.exceptions import AwsApiClientException
from phidata.utils.log import logger


class AwsApiClient(InfraApiClient):
    """This class is a wrapper around boto3 session to use with the AwsWorker"""

    def __init__(
        self,
        aws_region: Optional[str] = None,
        aws_profile: Optional[str] = None,
    ):

        super().__init__()
        # logger.debug("Creating AwsApiClient")
        # AWS configuration
        self.aws_region: Optional[str] = aws_region
        self.aws_profile: Optional[str] = aws_profile

        # aws boto3 session
        self._boto3_session: Optional[Any] = None
        logger.debug(f"**-+-** AwsApiClient created")

    def is_initialized(self) -> bool:
        if self.boto3_session is not None:
            return True
        return False

    def create_boto3_session(self) -> Optional[Any]:
        """Create a boto3 session"""
        from boto3 import session as boto3_session

        logger.debug(f"Creating boto3.Session")
        session = None
        try:
            session = boto3_session.Session(
                region_name=self.aws_region,
                profile_name=self.aws_profile,
            )
            logger.debug(f"**-+-** boto3.Session created")
            logger.debug(f"\taws_region: {session.region_name}")
            logger.debug(f"\taws_profile: {session.profile_name}")
        except Exception as e:
            raise

        if session is None:
            raise AwsApiClientException("Could not create boto3.Session")
        return session

    @property
    def boto3_session(self) -> Optional[Any]:
        if self._boto3_session is None:
            self._boto3_session = self.create_boto3_session()
        return self._boto3_session



from pathlib import Path
from typing import Optional, Any, Dict, List

from phidata.infra.aws.api_client import AwsApiClient
from phidata.infra.aws.resource.base import AwsResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class Secret(AwsResource):
    """
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html
    """

    resource_type = "Secret"
    service_name = "secretsmanager"

    # The name of the new secret.
    name: str
    client_request_token: Optional[str] = None
    # The description of the secret.
    description: Optional[str] = None
    kms_key_id: Optional[str] = None
    # The binary data to encrypt and store in the new version of the secret.
    # We recommend that you store your binary data in a file and then pass the contents of the file as a parameter.
    secret_binary: Optional[bytes] = None
    # The text data to encrypt and store in this new version of the secret.
    # We recommend you use a JSON structure of key/value pairs for your secret value.
    # Either SecretString or SecretBinary must have a value, but not both.
    secret_string: Optional[str] = None
    # A list of tags to attach to the secret.
    tags: Optional[List[Dict[str, str]]] = None
    # A list of Regions and KMS keys to replicate secrets.
    add_replica_regions: Optional[List[Dict[str, str]]] = None
    # Specifies whether to overwrite a secret with the same name in the destination Region.
    force_overwrite_replica_secret: Optional[str] = None

    # Read secret key/value pairs from yaml file
    read_from_yaml: Optional[Path] = None

    # provided by api on create
    secret_arn: Optional[str] = None
    secret_resource_name: Optional[str] = None
    secret_value: Optional[dict] = None

    def _create(self, aws_client: AwsApiClient) -> bool:
        """Creates the Secret

        Args:
            aws_client: The AwsApiClient for the current secret
        """
        print_info(f"Creating {self.get_resource_type()}: {self.get_resource_name()}")

        # Step 1: Build Secret configuration
        # create a dict of args which are not null, otherwise aws type validation fails
        not_null_args: Dict[str, Any] = {}
        if self.client_request_token:
            not_null_args["ClientRequestToken"] = self.client_request_token
        if self.description:
            not_null_args["Description"] = self.description
        if self.kms_key_id:
            not_null_args["KmsKeyId"] = self.kms_key_id
        if self.secret_binary:
            not_null_args["SecretBinary"] = self.secret_binary
        if self.secret_string:
            not_null_args["SecretString"] = self.secret_string
        if self.tags:
            not_null_args["Tags"] = self.tags
        if self.add_replica_regions:
            not_null_args["AddReplicaRegions"] = self.add_replica_regions
        if self.force_overwrite_replica_secret:
            not_null_args[
                "ForceOverwriteReplicaSecret"
            ] = self.force_overwrite_replica_secret

        # Step 2: Create Secret
        service_client = self.get_service_client(aws_client)
        try:
            created_resource = service_client.create_secret(
                Name=self.name,
                **not_null_args,
            )
            logger.debug(f"Secret: {created_resource}")

            # Validate Secret creation
            self.secret_arn = created_resource.get("ARN", None)
            self.secret_resource_name = created_resource.get("Name", None)
            logger.debug(f"secret_arn: {self.secret_arn}")
            logger.debug(f"secret_resource_name: {self.secret_resource_name}")
            if self.secret_arn is not None:
                print_info(f"Secret created: {self.name}")
                self.active_resource = created_resource
                self.save_resource_file()
                return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be created.")
            print_error(e)
        return False

    def _read(self, aws_client: AwsApiClient) -> Optional[Any]:
        """Returns the Secret

        Args:
            aws_client: The AwsApiClient for the current secret
        """
        logger.debug(f"Reading {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            describe_response = service_client.describe_secret(SecretId=self.name)
            logger.debug(f"Secret: {describe_response}")

            self.secret_arn = describe_response.get("ARN", None)
            self.secret_resource_name = describe_response.get("Name", None)
            logger.debug(f"secret_arn: {self.secret_arn}")
            logger.debug(f"secret_resource_name: {self.secret_resource_name}")
            if self.secret_arn is not None:
                print_info(f"Secret created: {self.name}")
                self.active_resource = describe_response
                self.save_resource_file()
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return self.active_resource

    def _delete(self, aws_client: AwsApiClient) -> bool:
        """Deletes the Secret

        Args:
            aws_client: The AwsApiClient for the current secret
        """
        print_info(f"Deleting {self.get_resource_type()}: {self.get_resource_name()}")

        service_client = self.get_service_client(aws_client)
        self.active_resource = None
        self.secret_value = None
        try:
            delete_response = service_client.delete_secret(SecretId=self.name)
            logger.debug(f"Secret: {delete_response}")
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} deleted"
            )
            self.save_resource_file()
            return True
        except Exception as e:
            print_error(f"{self.get_resource_type()} could not be deleted.")
            print_error("Please try again or delete resources manually.")
            print_error(e)
        return False

    def get_secret_value(self, aws_client: AwsApiClient) -> Any:
        """Get secret value

        Args:
            aws_client: The AwsApiClient for the current secret
        """
        logger.debug(f"Getting {self.get_resource_type()}: {self.get_resource_name()}")

        from botocore.exceptions import ClientError

        service_client = self.get_service_client(aws_client)
        try:
            secret_value = service_client.get_secret_value(SecretId=self.name)
            logger.debug(f"Secret: {secret_value}")

            if secret_value is None:
                logger.warning(f"Secret is None: {self.name}")
                return None

            self.secret_value = secret_value
            self.secret_arn = secret_value.get("ARN", None)
            self.secret_resource_name = secret_value.get("Name", None)
            self.save_resource_file()

            secret_string = secret_value.get("SecretString", None)
            if secret_string is not None:
                return secret_string

            secret_binary = secret_value.get("SecretBinary", None)
            if secret_binary is not None:
                return secret_binary
        except ClientError as ce:
            logger.debug(f"ClientError: {ce}")
        except Exception as e:
            print_error(f"Error reading {self.get_resource_type()}.")
            print_error(e)
        return None



from pathlib import Path
from typing import Optional, Dict, Any, Union, List, Tuple

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class WorkspaceVolumeType(ExtendedEnum):
    HostPath = "HostPath"
    EmptyDir = "EmptyDir"
    # PersistentVolume = "PersistentVolume"
    AwsEbs = "AwsEbs"
    # AwsEfs = "AwsEfs"


class PhidataAppArgs(PhidataBaseArgs):
    name: str

    # -*- Path parameters
    # The following args are populated by the K8sWorker and DockerWorker classes.
    # The build_resource_groups() function passes these args from the
    # WorkspaceConfig -> K8sConfig -> K8sArgs -> PhidataApp

    # Path to the workspace root directory
    workspace_root_path: Optional[Path] = None
    # Path to the workspace config file
    workspace_config_file_path: Optional[Path] = None
    # Path to important directories relative to the workspace_root
    # These directories are joined to the workspace_root dir
    #   to build paths depending on the environments (local, docker, k8s)
    # defaults are set by WorkspaceConfig.__init__()
    scripts_dir: Optional[str] = None
    storage_dir: Optional[str] = None
    meta_dir: Optional[str] = None
    products_dir: Optional[str] = None
    notebooks_dir: Optional[str] = None
    workflows_dir: Optional[str] = None
    workspace_config_dir: Optional[str] = None

    # -*- Common environment variables from the WorkspaceConfig
    # Env vars added to docker env when building PhidataApps
    #   and running workflows
    docker_env: Optional[Dict[str, str]] = None
    # Env vars added to k8s env when building PhidataApps
    #   and running workflows
    k8s_env: Optional[Dict[str, str]] = None

    # -*- Image Configuration
    image_name: Optional[str] = None
    image_tag: Optional[str] = None
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Install python dependencies using a requirements.txt file
    install_requirements: bool = False
    # Path to the requirements.txt file relative to the workspace_root
    requirements_file: str = "requirements.txt"

    # -*- Container Configuration
    # Each PhidataApp has 1 main container and multiple sidecar containers
    # The main container name
    container_name: Optional[str] = None
    # Overwrite the PYTHONPATH env var
    # which is usually set to the workspace_root_container_path
    python_path: Optional[str] = None
    # Add labels to the container
    container_labels: Optional[Dict[str, Any]] = None

    # Container env passed to the PhidataApp
    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None

    # Container secrets
    # Add secret variables to container env
    secrets: Optional[Dict[str, str]] = None
    # Read secret variables from a file in yaml format
    secrets_file: Optional[Path] = None
    # Read secret variables from AWS Secrets
    aws_secrets: Optional[Any] = None

    # Container ports
    # Open a container port if open_container_port=True
    open_container_port: bool = False
    # Port number on the container
    container_port: int = 8000
    # Port name: Only used by the K8sContainer
    container_port_name: str = "http"
    # Host port: Only used by the DockerContainer
    container_host_port: int = 8000

    # Container volumes
    # Mount the workspace directory on the container
    mount_workspace: bool = False
    workspace_volume_name: Optional[str] = None
    workspace_volume_type: Optional[WorkspaceVolumeType] = None
    # Path to mount the workspace volume
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace_root would be: /mnt/workspaces/idata
    workspace_volume_container_path: str = "/mnt/workspaces"
    # How to mount the workspace volume
    # Option 1: Mount the workspace from the host machine
    # If None, use the workspace_root_path
    # Note: This is the default on DockerContainers. We assume that DockerContainers
    # are running locally on the user's machine so the local workspace_root_path
    # is mounted to the workspace_volume_container_path
    workspace_volume_host_path: Optional[str] = None
    # Option 2: Load the workspace from git using a git-sync sidecar container
    # This the default on K8sContainers.
    create_git_sync_sidecar: bool = False
    # Required to create an initial copy of the workspace
    create_git_sync_init_container: bool = True
    git_sync_image_name: str = "k8s.gcr.io/git-sync"
    git_sync_image_tag: str = "v3.1.1"
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1

    # -*- Docker configuration
    # Run container in the background and return a Container object.
    container_detach: bool = True
    # Enable auto-removal of the container on daemon side when the container’s process exits.
    container_auto_remove: bool = True
    # Remove the container when it has finished running. Default: True.
    container_remove: bool = True
    # Username or UID to run commands as inside the container.
    container_user: Optional[Union[str, int]] = None
    # Keep STDIN open even if not attached.
    container_stdin_open: bool = True
    container_tty: bool = True
    # Specify a test to perform to check that the container is healthy.
    container_healthcheck: Optional[Dict[str, Any]] = None
    # Optional hostname for the container.
    container_hostname: Optional[str] = None
    # Platform in the format os[/arch[/variant]].
    container_platform: Optional[str] = None
    # Path to the working directory.
    container_working_dir: Optional[str] = None
    # Restart the container when it exits. Configured as a dictionary with keys:
    # Name: One of on-failure, or always.
    # MaximumRetryCount: Number of times to restart the container on failure.
    # For example: {"Name": "on-failure", "MaximumRetryCount": 5}
    container_restart_policy_docker: Optional[Dict[str, Any]] = None
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which adds the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes_docker: Optional[Dict[str, dict]] = None
    # Add ports to DockerContainer
    # The keys of the dictionary are the ports to bind inside the container,
    # either as an integer or a string in the form port/protocol, where the protocol is either tcp, udp.
    # The values of the dictionary are the corresponding ports to open on the host, which can be either:
    #   - The port number, as an integer.
    #       For example, {'2222/tcp': 3333} will expose port 2222 inside the container as port 3333 on the host.
    #   - None, to assign a random host port. For example, {'2222/tcp': None}.
    #   - A tuple of (address, port) if you want to specify the host interface.
    #       For example, {'1111/tcp': ('127.0.0.1', 1111)}.
    #   - A list of integers, if you want to bind multiple host ports to a single container port.
    #       For example, {'1111/tcp': [1234, 4567]}.
    container_ports_docker: Optional[Dict[str, Any]] = None

    # -*- K8s configuration
    # K8s Deployment configuration
    replicas: int = 1
    pod_name: Optional[str] = None
    deploy_name: Optional[str] = None
    secret_name: Optional[str] = None
    configmap_name: Optional[str] = None
    # Type: ImagePullPolicy
    image_pull_policy: Optional[Any] = None
    pod_annotations: Optional[Dict[str, str]] = None
    pod_node_selector: Optional[Dict[str, str]] = None
    # Type: RestartPolicy
    deploy_restart_policy: Optional[Any] = None
    deploy_labels: Optional[Dict[str, Any]] = None
    termination_grace_period_seconds: Optional[int] = None
    # How to spread the deployment across a topology
    # Key to spread the pods across
    topology_spread_key: Optional[str] = None
    # The degree to which pods may be unevenly distributed
    topology_spread_max_skew: Optional[int] = None
    # How to deal with a pod if it doesn't satisfy the spread constraint.
    topology_spread_when_unsatisfiable: Optional[str] = None

    # K8s Service Configuration
    create_service: bool = False
    service_name: Optional[str] = None
    # Type: ServiceType
    service_type: Optional[Any] = None
    # The port exposed by the service.
    service_port: int = 8000
    # The node_port exposed by the service if service_type = ServiceType.NODE_PORT
    service_node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    service_target_port: Optional[Union[str, int]] = None
    # Extra ports exposed by the webserver service. Type: List[CreatePort]
    service_ports: Optional[List[Any]] = None
    # Service labels
    service_labels: Optional[Dict[str, Any]] = None
    # Service annotations
    service_annotations: Optional[Dict[str, str]] = None
    # If ServiceType == ServiceType.LoadBalancer
    service_health_check_node_port: Optional[int] = None
    service_internal_traffic_policy: Optional[str] = None
    service_load_balancer_class: Optional[str] = None
    service_load_balancer_ip: Optional[str] = None
    service_load_balancer_source_ranges: Optional[List[str]] = None
    service_allocate_load_balancer_node_ports: Optional[bool] = None

    # K8s RBAC Configuration
    use_rbac: bool = False
    # Create a Namespace with name ns_name & default values
    ns_name: Optional[str] = None
    # or Provide the full Namespace definition
    # Type: CreateNamespace
    namespace: Optional[Any] = None
    # Create a ServiceAccount with name sa_name & default values
    sa_name: Optional[str] = None
    # or Provide the full ServiceAccount definition
    # Type: CreateServiceAccount
    service_account: Optional[Any] = None
    # Create a ClusterRole with name cr_name & default values
    cr_name: Optional[str] = None
    # or Provide the full ClusterRole definition
    # Type: CreateClusterRole
    cluster_role: Optional[Any] = None
    # Create a ClusterRoleBinding with name crb_name & default values
    crb_name: Optional[str] = None
    # or Provide the full ClusterRoleBinding definition
    # Type: CreateClusterRoleBinding
    cluster_role_binding: Optional[Any] = None

    # Add additional Kubernetes resources to the App
    # Type: CreateSecret
    extra_secrets: Optional[List[Any]] = None
    # Type: CreateConfigMap
    extra_configmaps: Optional[List[Any]] = None
    # Type: CreateService
    extra_services: Optional[List[Any]] = None
    # Type: CreateDeployment
    extra_deployments: Optional[List[Any]] = None
    # Type: CreatePersistentVolume
    extra_pvs: Optional[List[Any]] = None
    # Type: CreatePVC
    extra_pvcs: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_containers: Optional[List[Any]] = None
    # Type: CreateContainer
    extra_init_containers: Optional[List[Any]] = None
    # Type: CreatePort
    extra_ports: Optional[List[Any]] = None
    # Type: CreateVolume
    extra_volumes: Optional[List[Any]] = None
    # Type: CreateStorageClass
    extra_storage_classes: Optional[List[Any]] = None
    # Type: CreateCustomObject
    extra_custom_objects: Optional[List[Any]] = None
    # Type: CreateCustomResourceDefinition
    extra_crds: Optional[List[Any]] = None

    # Other args
    print_env_on_load: bool = True
    # If True, skip resource creation if active resources with the same name exist.
    use_cache: bool = True

    # Extra kwargs used to ensure older versions of phidata don't throw syntax errors
    extra_kwargs: Optional[Dict[str, Any]] = None

    # -*- AWS parameters
    # Common aws params used by apps, resources and data assets
    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None
    aws_config_file: Optional[str] = None
    aws_shared_credentials_file: Optional[str] = None

    class Config:
        arbitrary_types_allowed = True


class PhidataApp(PhidataBase):
    """Base Class for all PhidataApps"""

    def __init__(self) -> None:
        super().__init__()

        # Cache env_data & secret_data
        self.env_data: Optional[Dict[str, Any]] = None
        self.secret_data: Optional[Dict[str, Any]] = None

        # Args for the PhidataApp, provided by the subclass
        self.args: PhidataAppArgs

        # Dict of DockerResourceGroups
        # Type: Optional[Dict[str, DockerResourceGroup]]
        self.docker_resource_groups: Optional[Dict[str, Any]] = None

        # Dict of KubernetesResourceGroups
        # Type: Optional[Dict[str, K8sResourceGroup]]
        self.k8s_resource_groups: Optional[Dict[str, Any]] = None

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path if self.args else None

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir if self.args else None

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir if self.args else None

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir if self.args else None

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir if self.args else None

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir if self.args else None

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workflows_dir(self) -> Optional[str]:
        return self.args.workflows_dir if self.args else None

    @workflows_dir.setter
    def workflows_dir(self, workflows_dir: str) -> None:
        if self.args is not None and workflows_dir is not None:
            self.args.workflows_dir = workflows_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir if self.args else None

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env if self.args else None

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env if self.args else None

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region if self.args else None

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile if self.args else None

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file if self.args else None

    @aws_config_file.setter
    def aws_config_file(self, aws_config_file: str) -> None:
        if self.args is not None and aws_config_file is not None:
            self.args.aws_config_file = aws_config_file

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file if self.args else None

    @aws_shared_credentials_file.setter
    def aws_shared_credentials_file(self, aws_shared_credentials_file: str) -> None:
        if self.args is not None and aws_shared_credentials_file is not None:
            self.args.aws_shared_credentials_file = aws_shared_credentials_file

    ######################################################
    ## Get App Attributes
    ######################################################

    def get_image_str(self):
        return f"{self.args.image_name}:{self.args.image_tag}"

    def get_container_name(self) -> str:
        from phidata.utils.common import get_default_container_name

        return self.args.container_name or get_default_container_name(self.args.name)

    def get_container_port(self) -> int:
        return self.args.container_port

    def get_container_host_port(self) -> int:
        return self.args.container_host_port

    def get_pod_name(self) -> str:
        from phidata.utils.common import get_default_pod_name

        return self.args.pod_name or get_default_pod_name(self.args.name)

    def get_deploy_name(self) -> str:
        from phidata.utils.common import get_default_deploy_name

        return self.args.deploy_name or get_default_deploy_name(self.args.name)

    def get_secret_name(self) -> str:
        from phidata.utils.common import get_default_secret_name

        return self.args.secret_name or get_default_secret_name(self.args.name)

    def get_configmap_name(self) -> str:
        from phidata.utils.common import get_default_configmap_name

        return self.args.configmap_name or get_default_configmap_name(self.args.name)

    def get_service_name(self) -> str:
        from phidata.utils.common import get_default_service_name

        return self.args.service_name or get_default_service_name(self.args.name)

    def get_service_port(self) -> int:
        return self.args.service_port

    def get_sa_name(self) -> str:
        from phidata.utils.common import get_default_sa_name

        return self.args.sa_name or get_default_sa_name(self.args.name)

    def get_cr_name(self) -> str:
        from phidata.utils.common import get_default_cr_name

        return self.args.cr_name or get_default_cr_name(self.args.name)

    def get_crb_name(self) -> str:
        from phidata.utils.common import get_default_crb_name

        return self.args.crb_name or get_default_crb_name(self.args.name)

    def get_env_data(self) -> Optional[Dict[str, str]]:
        if self.env_data is None:
            self.env_data = self.read_yaml_file(file_path=self.args.env_file)
        return self.env_data

    def get_secret_data(self) -> Optional[Dict[str, str]]:
        if self.secret_data is None:
            self.secret_data = self.read_yaml_file(file_path=self.args.secrets_file)
        return self.secret_data

    ######################################################
    ## Docker functions
    ######################################################

    def init_docker_resource_groups(self, docker_build_context: Any) -> None:
        logger.debug(
            f"@init_docker_resource_groups not defined for {self.__class__.__name__}"
        )

    def get_docker_resource_groups(
        self, docker_build_context: Any
    ) -> Optional[Dict[str, Any]]:
        if self.docker_resource_groups is None:
            self.init_docker_resource_groups(docker_build_context)
        # # Comment out in production
        # if self.docker_resource_groups:
        #     logger.debug("DockerResourceGroups:")
        #     for rg_name, rg in self.docker_resource_groups.items():
        #         logger.debug(
        #             "{}: {}".format(rg_name, rg.json(exclude_none=True, indent=2))
        #         )
        return self.docker_resource_groups

    ######################################################
    ## K8s functions
    ######################################################

    def init_k8s_resource_groups(self, k8s_build_context: Any) -> None:
        logger.debug(
            f"@init_docker_resource_groups not defined for {self.__class__.__name__}"
        )

    def get_k8s_resource_groups(
        self, k8s_build_context: Any
    ) -> Optional[Dict[str, Any]]:
        if self.k8s_resource_groups is None:
            self.init_k8s_resource_groups(k8s_build_context)
        # # Comment out in production
        # if self.k8s_resource_groups:
        #     logger.debug("K8sResourceGroups:")
        #     for rg_name, rg in self.k8s_resource_groups.items():
        #         logger.debug(
        #             "{}:{}\n{}".format(rg_name, type(rg), rg)
        #         )
        return self.k8s_resource_groups

    ######################################################
    ## Helpers
    ######################################################

    def get_container_paths(self) -> Optional[Any]:
        if self.workspace_root_path is None:
            return None

        workspace_name = self.workspace_root_path.stem
        if workspace_name is None:
            return None

        workspace_volume_container_path: str = self.args.workspace_volume_container_path
        if workspace_volume_container_path is None:
            return None

        from phidata.types.context import ContainerPathContext

        workspace_root_container_path = (
            f"{workspace_volume_container_path}/{workspace_name}"
        )
        container_paths = ContainerPathContext(
            workspace_name=workspace_name,
            workspace_root=workspace_root_container_path,
            workspace_parent=workspace_volume_container_path,
        )

        if self.args.scripts_dir is not None:
            container_paths.scripts_dir = (
                f"{workspace_root_container_path}/{self.args.scripts_dir}"
            )

        if self.args.storage_dir is not None:
            container_paths.storage_dir = (
                f"{workspace_root_container_path}/{self.args.storage_dir}"
            )

        if self.args.meta_dir is not None:
            container_paths.meta_dir = (
                f"{workspace_root_container_path}/{self.args.meta_dir}"
            )

        if self.args.products_dir is not None:
            container_paths.products_dir = (
                f"{workspace_root_container_path}/{self.args.products_dir}"
            )

        if self.args.notebooks_dir is not None:
            container_paths.notebooks_dir = (
                f"{workspace_root_container_path}/{self.args.notebooks_dir}"
            )

        if self.args.workflows_dir is not None:
            container_paths.workflows_dir = (
                f"{workspace_root_container_path}/{self.args.workflows_dir}"
            )

        if self.args.workspace_config_dir is not None:
            container_paths.workspace_config_dir = (
                f"{workspace_root_container_path}/{self.args.workspace_config_dir}"
            )

        if self.args.requirements_file is not None:
            container_paths.requirements_file = (
                f"{workspace_root_container_path}/{self.args.requirements_file}"
            )

        return container_paths

    def read_yaml_file(self, file_path: Optional[Path]) -> Optional[Dict[str, Any]]:
        if file_path is not None and file_path.exists() and file_path.is_file():
            import yaml

            # logger.debug(f"Reading {file_path}")
            data_from_file = yaml.safe_load(file_path.read_text())
            if data_from_file is not None and isinstance(data_from_file, dict):
                return data_from_file
            else:
                logger.error(f"Invalid file: {file_path}")
        return None

    def set_aws_env_vars(self, env_dict: Dict[str, str]) -> None:
        """Set AWS environment variables."""
        from phidata.constants import (
            AWS_REGION_ENV_VAR,
            AWS_DEFAULT_REGION_ENV_VAR,
            # AWS_PROFILE_ENV_VAR,
            # AWS_CONFIG_FILE_ENV_VAR,
            # AWS_SHARED_CREDENTIALS_FILE_ENV_VAR,
        )

        if self.aws_region is not None:
            env_dict[AWS_REGION_ENV_VAR] = self.aws_region
            env_dict[AWS_DEFAULT_REGION_ENV_VAR] = self.aws_region
        # if self.aws_profile is not None:
        #     env_dict[AWS_PROFILE_ENV_VAR] = self.aws_profile
        # if self.aws_config_file is not None:
        #     env_dict[AWS_CONFIG_FILE_ENV_VAR] = self.aws_config_file
        # if self.aws_shared_credentials_file is not None:
        #     env_dict[
        #         AWS_SHARED_CREDENTIALS_FILE_ENV_VAR
        #     ] = self.aws_shared_credentials_file



from collections import OrderedDict
from typing import Dict, List, Type, Union

from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.image import DockerImage
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.volume import DockerVolume
from phidata.infra.docker.resource.base import DockerResource

# Use this as a type for an object which can hold any DockerResource
DockerResourceType = Union[
    DockerNetwork,
    DockerImage,
    DockerVolume,
    DockerContainer,
]

# Use this as an ordered list to iterate over all DockerResource Classes
# This list is the order in which resources are installed as well.
DockerResourceTypeList: List[Type[DockerResource]] = [
    DockerNetwork,
    DockerImage,
    DockerVolume,
    DockerContainer,
]

# Maps each DockerResource to an install weight
# lower weight DockerResource(s) get installed first
# i.e. networks are installed first, then volumes ... and so on
DockerResourceInstallOrder: Dict[str, int] = OrderedDict(
    {
        resource_type.__name__: idx
        for idx, resource_type in enumerate(DockerResourceTypeList, start=1)
    }
)



from typing import Optional, Any, Dict, List

from phidata.infra.docker.api_client import DockerApiClient
from phidata.infra.docker.resource.base import DockerResource
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class DockerImage(DockerResource):
    resource_type: str = "Image"

    # Docker image name, usually as repo/image
    name: str
    # Docker image tag
    tag: Optional[str] = None

    # Path to the directory containing the Dockerfile
    path: Optional[str] = None
    # Path to the Dockerfile within the build context
    dockerfile: Optional[str] = None

    # Print the build log
    print_build_log: bool = False
    # Push the image to the registry. Similar to the docker push command.
    push_image: bool = False
    print_push_output: bool = False

    # Remove intermediate containers.
    # The docker build command defaults to --rm=true,
    # The docker api kept the old default of False to preserve backward compatibility
    rm: Optional[bool] = True
    # Always remove intermediate containers, even after unsuccessful builds
    forcerm: Optional[bool] = None
    # HTTP timeout
    timeout: Optional[int] = None
    # Downloads any updates to the FROM image in Dockerfiles
    pull: Optional[bool] = None
    # Skips docker cache when set to True
    # i.e. rebuilds all layes of the image
    skip_docker_cache: Optional[bool] = None
    # A dictionary of build arguments
    buildargs: Optional[Dict[str, Any]] = None
    # A dictionary of limits applied to each container created by the build process. Valid keys:
    # memory (int): set memory limit for build
    # memswap (int): Total memory (memory + swap), -1 to disable swap
    # cpushares (int): CPU shares (relative weight)
    # cpusetcpus (str): CPUs in which to allow execution, e.g. "0-3", "0,1"
    container_limits: Optional[Dict[str, Any]] = None
    # Size of /dev/shm in bytes. The size must be greater than 0. If omitted the system uses 64MB
    shmsize: Optional[int] = None
    # A dictionary of labels to set on the image
    labels: Optional[Dict[str, Any]] = None
    # A list of images used for build cache resolution
    cache_from: Optional[List[Any]] = None
    # Name of the build-stage to build in a multi-stage Dockerfile
    target: Optional[str] = None
    # networking mode for the run commands during build
    network_mode: Optional[str] = None
    # Squash the resulting images layers into a single layer.
    squash: Optional[bool] = None
    # Extra hosts to add to /etc/hosts in building containers, as a mapping of hostname to IP address.
    extra_hosts: Optional[Dict[str, Any]] = None
    # Platform in the format os[/arch[/variant]].
    platform: Optional[str] = None
    # Isolation technology used during build. Default: None.
    isolation: Optional[str] = None
    # If True, and if the docker client configuration file (~/.docker/config.json by default)
    # contains a proxy configuration, the corresponding environment variables
    # will be set in the container being built.
    use_config_proxy: Optional[bool] = None

    # so that images arent deleted when phi ws down is run
    skip_delete: bool = True
    image_build_id: Optional[str] = None

    def get_resource_name(self) -> Optional[str]:
        return self.get_name_tag()

    def get_name_tag(self) -> str:
        image_name_tag = self.name
        if self.tag:
            image_name_tag = f"{image_name_tag}:{self.tag}"
        return image_name_tag

    def build_image(self, docker_client: DockerApiClient) -> Optional[Any]:
        from docker import DockerClient
        from docker.errors import BuildError, APIError

        print_info(f"Building image: {self.get_name_tag()}")
        nocache = self.skip_docker_cache
        pull = self.pull
        if self.path is not None:
            print_info(f"\t  path: {self.path}")
        if self.dockerfile is not None:
            print_info(f"    dockerfile: {self.dockerfile}")
        logger.debug(f"platform: {self.platform}")
        logger.debug(f"nocache: {nocache}")
        logger.debug(f"pull: {pull}")
        try:
            _api_client: DockerClient = docker_client.api_client
            build_stream = _api_client.api.build(
                tag=self.get_name_tag(),
                path=self.path,
                dockerfile=self.dockerfile,
                nocache=nocache,
                rm=self.rm,
                forcerm=self.forcerm,
                timeout=self.timeout,
                pull=pull,
                buildargs=self.buildargs,
                container_limits=self.container_limits,
                shmsize=self.shmsize,
                labels=self.labels,
                cache_from=self.cache_from,
                target=self.target,
                network_mode=self.network_mode,
                squash=self.squash,
                extra_hosts=self.extra_hosts,
                platform=self.platform,
                isolation=self.isolation,
                use_config_proxy=self.use_config_proxy,
                decode=True,
            )
            for build_log in build_stream:
                line = build_log.get("stream", None)
                if line is None:
                    continue
                if len(line) < 5:
                    continue
                line = line.strip()
                if "Step" in line and self.print_build_log:
                    print_info(line)
                else:
                    logger.debug(line)
                if "ERROR" in line or "error" in line:
                    print_error(f"Image build failed: {self.get_name_tag()}")
                    print_error(line)
                    return None
                if build_log.get("aux", None) is not None:
                    logger.debug("build_log['aux'] :{}".format(build_log["aux"]))
                    self.image_build_id = build_log.get("aux", {}).get("ID")
            if self.push_image:
                print_info(f"Pushing {self.get_name_tag()}")
                push_progress = None
                prev_push_progress = None
                for push_output in _api_client.images.push(
                    repository=self.name,
                    tag=self.tag,
                    stream=True,
                    decode=True,
                ):
                    if push_output.get("error", None) is not None:
                        print_error(push_output["error"])
                        print_error(f"Push failed for {self.get_name_tag()}")
                        print_error(
                            "If you are using a private registry, make sure you are logged in"
                        )
                    if self.print_push_output and push_output.get("status", None) in (
                        "Pushing",
                        "Pushed",
                    ):
                        push_progress = push_output.get("progress", None)
                        if push_progress != prev_push_progress:
                            print_info(push_progress)
                            prev_push_progress = push_progress
                    if push_output.get("aux", {}).get("Size", 0) > 0:
                        print_info(f"Push complete: {push_output.get('aux', {})}")
            return self._read(docker_client)
        except TypeError as type_error:
            print_error(type_error)
            # print_error("TypeError: {}".format(type_error))
        except BuildError as build_error:
            print_error(build_error)
            # print_error("BuildError: {}".format(build_error))
        except APIError as api_err:
            print_error(api_err)
            # print_error("ApiError: {}".format(api_err))
        return None

    def _create(self, docker_client: DockerApiClient) -> bool:
        """Creates the image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker.models.images import Image

        logger.debug("Creating: {}".format(self.get_resource_name()))

        try:
            image_object = self.build_image(docker_client)
            if image_object is not None and isinstance(image_object, Image):
                logger.debug("Image built: {}".format(image_object))
                self.active_resource = image_object
                self.active_resource_class = Image
                return True
            else:
                logger.error("Image {} could not be built".format(self.tag))
        except Exception as e:
            logger.exception(e)
            logger.error("Error while creating image: {}".format(e))
            raise

        return False

    def _read(self, docker_client: DockerApiClient) -> Any:
        """Returns an Image object if available"""
        from docker import DockerClient
        from docker.models.images import Image
        from docker.errors import ImageNotFound, NotFound

        logger.debug("Reading: {}".format(self.get_name_tag()))
        try:
            _api_client: DockerClient = docker_client.api_client
            image_object: Optional[List[Image]] = _api_client.images.get(
                name=self.get_name_tag()
            )
            if image_object is not None and isinstance(image_object, Image):
                logger.debug("Image found: {}".format(image_object))
                self.active_resource = image_object
                return image_object
        except (NotFound, ImageNotFound) as not_found_err:
            logger.debug(f"Image {self.tag} not found")

        return None

    def _update(self, docker_client: DockerApiClient) -> bool:
        """Updates the Image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        logger.debug("Updating: {}".format(self.get_resource_name()))
        return self._create(docker_client=docker_client)

    def _delete(self, docker_client: DockerApiClient) -> bool:
        """Deletes the Image

        Args:
            docker_client: The DockerApiClient for the current cluster
        """
        from docker import DockerClient
        from docker.models.images import Image

        logger.debug("Deleting: {}".format(self.get_resource_name()))
        image_object: Optional[Image] = self._read(docker_client)
        # Return True if there is no image to delete
        if image_object is None:
            logger.debug("No image to delete")
            return True

        # Delete Image
        try:
            self.active_resource = None
            logger.debug("Deleting image: {}".format(self.tag))
            _api_client: DockerClient = docker_client.api_client
            _api_client.images.remove(image=self.tag, force=True)
            return True
        except Exception as e:
            logger.exception("Error while deleting image: {}".format(e))

        return False



from pathlib import Path
from typing import Optional, Literal, Union, List

from phidata.asset.file import File, FileType
from phidata.asset.table.sql import SqlTable
from phidata.utils.cli_console import print_info, print_warning
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class DownloadUrlToSqlArgs(PythonTaskArgs):
    url: str
    sql_table: SqlTable
    # How to behave if the sql_table already exists.
    # fail: Raise a ValueError.
    # replace: Drop the table before inserting new values.
    # append: Insert new values to the existing table.
    if_exists: Optional[Literal["fail", "replace", "append"]] = None
    # Write DataFrame index as a column.
    # Uses index_label as the column name in the table.
    index: Optional[bool] = None
    # Column label for index column(s).
    # If None is given (default) and index is True, then the index names are used.
    # A sequence should be given if the DataFrame uses MultiIndex.
    index_label: Optional[Union[str, List[str]]] = None
    # Specify the number of rows in each batch to be written at a time.
    # By default, all rows will be written at once.
    chunksize: Optional[int] = None
    create_db_engine_from_conn_id: bool = True
    # Optional File object to use
    file: Optional[File] = None


class DownloadUrlToSql(PythonTask):
    def __init__(
        self,
        url: str,
        sql_table: SqlTable,
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        index: Optional[bool] = None,
        index_label: Optional[Union[str, List[str]]] = None,
        chunksize: Optional[int] = None,
        file: Optional[File] = None,
        name: str = "download_url_to_sql",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: DownloadUrlToSqlArgs = DownloadUrlToSqlArgs(
                url=url,
                sql_table=sql_table,
                if_exists=if_exists,
                index=index,
                index_label=index_label,
                chunksize=chunksize,
                file=file,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=download_url_to_sql,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def url(self) -> str:
        return self.args.url

    @url.setter
    def url(self, url: str) -> None:
        if url is not None:
            self.args.url = url

    @property
    def sql_table(self) -> Optional[SqlTable]:
        return self.args.sql_table

    @sql_table.setter
    def sql_table(self, sql_table: SqlTable) -> None:
        if sql_table is not None:
            self.args.sql_table = sql_table

    @property
    def file(self) -> Optional[File]:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file


def best_guess_file_name(url: str) -> str:

    from phidata.utils.dttm import get_today_utc_date_str

    fn: Optional[str] = None
    if "/" in url:
        fn = url.split("/")[-1]
    if fn is None:
        fn = f"tmp__{get_today_utc_date_str()}"
    logger.debug("fn: {}".format(fn))
    return fn


def download_url_to_file(args: DownloadUrlToSqlArgs) -> bool:

    import httpx
    import rich.progress

    url = args.url
    if url is None:
        logger.error("URL not available")
        return False

    if args.file is None:
        args.file = File.temporary(best_guess_file_name(args.url))

    file_path: Optional[Path] = args.file.file_path
    if file_path is None:
        logger.error("FilePath not available")
        return False
    file_dir: Path = file_path.parent.resolve()

    # Download URL contents
    print_info("Downloading")
    print_info(f"  Url: {url}")
    print_info(f"  To : {file_path}")

    # Create the directory if it does not exist
    if not file_dir.exists():
        file_dir.mkdir(parents=True, exist_ok=True)

    bytes_to_download: int = 0
    try:
        with file_path.open("w") as open_file:
            with httpx.stream("GET", url) as response:
                try:
                    headers: httpx.Headers = response.headers
                    # logger.debug("headers: {}".format(headers))
                    # logger.debug("status_code: {}".format(response.status_code))
                    bytes_to_download = int(headers["Content-Length"])
                    # logger.debug("bytes_to_download : {}".format(bytes_to_download))
                except Exception:
                    print_warning(f"Could not get total bytes_to_download from headers")

                with rich.progress.Progress(
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    rich.progress.BarColumn(50),
                    rich.progress.DownloadColumn(),
                    rich.progress.TransferSpeedColumn(),
                ) as progress:
                    download_task = progress.add_task(
                        "Download", total=bytes_to_download
                    )
                    for chunk in response.iter_text():
                        open_file.write(chunk)
                        progress.update(
                            download_task, completed=response.num_bytes_downloaded
                        )
        logger.info("Url downloaded")
        return True
    except Exception as e:
        logger.error("Could not download url, please try again")
        logger.exception(e)
    return False


def load_file_to_sql_table_pandas(args: DownloadUrlToSqlArgs) -> bool:

    import pandas as pd

    file_to_load: Optional[File] = args.file
    if file_to_load is None:
        logger.error("File not available")
        return False
    file_path: Optional[Path] = file_to_load.file_path
    if file_path is None:
        logger.error("FilePath not available")
        return False
    file_type: Optional[FileType] = file_to_load.file_type
    if file_type is None:
        logger.error("FileType not available")
        return False
    # logger.debug("File: {}".format(file_to_load.args))

    sql_table: Optional[SqlTable] = args.sql_table
    if sql_table is None:
        logger.error("SqlTable not available")
        return False
    # logger.debug("SqlTable: {}".format(sql_table.args))

    logger.info("Reading: {}".format(file_path))
    df: Optional[pd.DataFrame] = None
    if file_type == FileType.CSV:
        df = pd.read_csv(file_path)
    elif file_type == FileType.JSON:
        df = pd.read_json(file_path)
    elif file_type == FileType.TSV:
        df = pd.read_csv(file_path, sep="\t")

    if df is not None:
        # logger.info()("DataFrame:\n{}".format(df.head()))
        logger.info("Writing to table: {}".format(sql_table.name))
        upload_success = sql_table.write_pandas_df(
            df=df,
            if_exists=args.if_exists,
            index=args.index,
            index_label=args.index_label,
            chunksize=args.chunksize,
        )
        return upload_success
    else:
        logger.info("Could not read file into DataFrame")
        return False


def download_url_to_sql(**kwargs) -> bool:

    args: DownloadUrlToSqlArgs = DownloadUrlToSqlArgs.from_kwargs(kwargs)
    # logger.info(f"DownloadUrlToSqlArgs: {args}")

    download_success = download_url_to_file(args)
    if download_success:
        return load_file_to_sql_table_pandas(args)
    return False
